################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# get_host_interfaces offers similar functionality as netifaces

from ctwin32.iphlpapi import (
    get_host_interfaces,
    ConvertInterfaceGuidToLuid,
    ConvertInterfaceLuidToAlias,
    ConvertInterfaceIndexToLuid,
    )

print("\nAdapters by index:\n")
idx = 1
while True:
    try:
        alias = ConvertInterfaceLuidToAlias(ConvertInterfaceIndexToLuid(idx))
        print(f"{idx:3}: {alias}")
        idx += 1
    except OSError:
        break

print("\n\nInterfaces by adapter:")
for guid, ifaces in get_host_interfaces(None, True).items():
    alias = ConvertInterfaceLuidToAlias(ConvertInterfaceGuidToLuid(guid))
    print(f"\nAdapter: {alias}, {guid}")
    for nif in ifaces:
        print(f"    {nif!r}")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This sample demonstrates password-less logons via 'S4U' (service for user).
#
################################################################################

import sys
import ctypes
from ctwin32 import advapi, kernel, secur, user
from ctwin32 import (
    CREATE_NEW_CONSOLE,
    MAXIMUM_ALLOWED,
    Network,
    SE_ASSIGNPRIMARYTOKEN_PRIVILEGE,
    SE_GROUP_ENABLED,
    SE_GROUP_ENABLED_BY_DEFAULT,
    SE_GROUP_LOGON_ID,
    SE_IMPERSONATE_PRIVILEGE,
    SE_TCB_PRIVILEGE,
    SecurityImpersonation,
    TOKEN_DUPLICATE,
    TokenImpersonation,
    UOI_USER_SID,
    WinInteractiveSid,
    WinRestrictedCodeSid,
    )
from ctwin32.wtypes import (
    WCHAR,
    wchar_len_sz,
    WCHAR_SIZE,
    ULONG,
    UNICODE_STRING,
    )

################################################################################

def enable_s4u_privileges():
    required = [SE_TCB_PRIVILEGE, SE_ASSIGNPRIMARYTOKEN_PRIVILEGE]
    try:
        # either we are running in an highly privileged process (e.g. SYSTEM)
        # where we can simply enable the requires privileges...
        advapi.enable_privileges(required)
    except OSError:
        pass
    else:
        return

    # ... or we need to run in a process that is somewhat privileged (e.g.
    # administrator) so we can fetch and impersonate the SYSTEM identity
    advapi.enable_privileges([SE_IMPERSONATE_PRIVILEGE])
    session = kernel.ProcessIdToSessionId(kernel.GetCurrentProcessId())
    with advapi.open_system_token_for_session(TOKEN_DUPLICATE, session) as tok:
        dup_tok_args = (
            tok,
            MAXIMUM_ALLOWED,
            kernel.SECURITY_ATTRIBUTES(),
            SecurityImpersonation,
            TokenImpersonation,
            )
        with advapi.DuplicateTokenEx(*dup_tok_args) as dup:
            advapi.enable_token_privileges(dup, required)
            advapi.SetThreadToken(dup)

################################################################################

def kerb_s4u_logon(upn, realm, flags=0):
    """Create a self contained KERB_S4U_LOGON structure as required for
    calling LsaLogonUser().
    """
    wlu = wchar_len_sz(upn)     # aka 'user name'
    wlr = wchar_len_sz(realm)   # aka 'domain name'

    class KERB_S4U_LOGON(ctypes.Structure):
        KerbS4ULogon = 12
        _fields_ = (
            ("MessageType", ULONG),
            ("Flags", ULONG),
            ("ClientUpn", UNICODE_STRING),
            ("ClientRealm", UNICODE_STRING),
            ("upn_buf", WCHAR * wlu),
            ("realm_buf", WCHAR * wlr),
            )

        def __init__(self, upn, realm, flags):
            blu = wlu * WCHAR_SIZE
            blr = wlr * WCHAR_SIZE
            addr = ctypes.addressof(self)
            super().__init__(
                self.KerbS4ULogon,
                flags,
                (blu - WCHAR_SIZE, blu, addr + __class__.upn_buf.offset),
                (blr - WCHAR_SIZE, blr, addr + __class__.realm_buf.offset),
                upn,
                realm
                )

    return KERB_S4U_LOGON(upn, realm, flags)

################################################################################

def get_additional_groups():
    attr2 = SE_GROUP_ENABLED_BY_DEFAULT | SE_GROUP_ENABLED
    attr1 = attr2 | SE_GROUP_LOGON_ID
    sid1 = user.GetUserObjectInformation(
        user.GetProcessWindowStation(),
        UOI_USER_SID
        ).raw
    if not sid1:
        # most likely running in session 0
        sid1 = advapi.CreateWellKnownSid(WinRestrictedCodeSid)
        attr1= attr2
    sid2 = advapi.CreateWellKnownSid(WinInteractiveSid)
    return advapi.make_token_groups(((sid1, attr1), (sid2, attr2)))

################################################################################

def logon_s4u(user, domain, add_groups):
    lsa = secur.LsaConnectUntrusted()
    try:
        auth_pkg = secur.LsaLookupAuthenticationPackage(lsa, "Negotiate")
        ks4u = kerb_s4u_logon(user, domain)
        return secur.LsaLogonUser(
            lsa,
            "run_s4u",
            Network,
            auth_pkg,
            ks4u,
            add_groups,
            "run_s4u",
            )
    finally:
        secur.LsaDeregisterLogonProcess(lsa)

################################################################################

if __name__ == "__main__":
    try:
        enable_s4u_privileges()
    except OSError:
        print("Not enough privileges - try to run as administrator.")
        sys.exit(1)

    name = sys.argv[1]
    lst = name.split("\\", 1)
    domain, name = lst if len(lst) > 1 else (".", name)

    token, _ = logon_s4u(name, domain, get_additional_groups())
    advapi.create_process_as_user(
        token,
        ["cmd.exe", "/K", "whoami /user"],
        CREATE_NEW_CONSOLE
        )

################################################################################




################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This example shows how you can use a job object to collect information
# about a process.

import sys
from ctwin32 import (
    kernel,
    CREATE_SUSPENDED,
    INFINITE,
    JobObjectBasicAccountingInformation,
    )

################################################################################

def fmt_ft(ft):
    q, _ = divmod(int(ft), 10000)
    q, ms = divmod(q, 1000)
    q, s = divmod(q, 60)
    h, m = divmod(q, 60)
    return f"{h}:{m:02}:{s:02}.{ms:03}"

################################################################################

if __name__ == "__main__":

    arglist = ["cmd", "/c", *sys.argv[1:]]
    with kernel.CreateJobObject() as job:
        with kernel.create_process(arglist, CREATE_SUSPENDED) as pi:
            kernel.AssignProcessToJobObject(job, pi.hProcess)
            create_time = kernel.GetSystemTimeAsFileTime()
            kernel.ResumeThread(pi.hThread)
            kernel.WaitForSingleObject(pi.hProcess, INFINITE)
            exit_time = kernel.GetSystemTimeAsFileTime()

        info_cls = JobObjectBasicAccountingInformation
        info = kernel.JOBOBJECT_BASIC_ACCOUNTING_INFORMATION()
        kernel.QueryInformationJobObject(job, info_cls, info)
        print(f"\n\nduration  : {fmt_ft(exit_time - create_time)}")
        print(f"kernel    : {fmt_ft(info.TotalKernelTime)}")
        print(f"user      : {fmt_ft(info.TotalUserTime)}")
        print(f"processes : {info.TotalProcesses - 1}")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    DWORD,
    GUID,
    INT,
    PDWORD,
    PGUID,
    PINT,
    PVOID,
    PWSTR,
    PULONG,
    ULONG,
    WinError,
    )
from . import (
    ref,
    fun_fact,
    multi_str_from_str,
    MAX_DEVICE_ID_LEN,
    MAX_PATH,
    CM_GET_DEVICE_INTERFACE_LIST_PRESENT,
    CM_LOCATE_DEVNODE_NORMAL,
    CR_BUFFER_SMALL,
    CR_SUCCESS,
    ERROR_FLOPPY_UNKNOWN_ERROR,
    ERROR_CURRENT_DIRECTORY,
    )
from .advapi import registry_to_py
_cfg = ctypes.WinDLL("cfgmgr32.dll", use_last_error=True)

################################################################################

CM_MapCrToWin32Err = fun_fact(_cfg.CM_MapCrToWin32Err, (DWORD, DWORD, DWORD))

def raise_on_cr(cfgret):
    if cfgret != CR_SUCCESS:
        # use an unlikely default value. anyone out there still using floppies?
        err = CM_MapCrToWin32Err(cfgret, ERROR_FLOPPY_UNKNOWN_ERROR)
        raise WinError(err)

################################################################################

_CM_Get_Device_ID = fun_fact(
    _cfg.CM_Get_Device_IDW,
    (DWORD, DWORD, PWSTR, ULONG, ULONG)
    )

def CM_Get_Device_ID(devinst):
    idstr = string_buffer(MAX_DEVICE_ID_LEN)
    raise_on_cr(_CM_Get_Device_ID(devinst, idstr, MAX_DEVICE_ID_LEN, 0))
    return idstr.value

################################################################################

_CM_Get_DevNode_Status = fun_fact(
    _cfg.CM_Get_DevNode_Status,
    (DWORD, PULONG, PULONG, DWORD, ULONG)
    )

def CM_Get_DevNode_Status(dev_inst):
    status = ULONG()
    problem = ULONG()
    raise_on_cr(_CM_Get_DevNode_Status(ref(status), ref(problem), dev_inst, 0))
    return status.value, problem.value

################################################################################

_CM_Enumerate_Enumerators = fun_fact(
    _cfg.CM_Enumerate_EnumeratorsW,
    (DWORD, ULONG, PWSTR, PULONG, ULONG)
    )

def CM_Enumerate_Enumerators(idx):
    enum_str = string_buffer(MAX_DEVICE_ID_LEN)
    size = ULONG(MAX_DEVICE_ID_LEN)
    raise_on_cr(_CM_Enumerate_Enumerators(idx, enum_str, ref(size), 0))
    return enum_str.value

################################################################################

_CM_Enumerate_Classes = fun_fact(
    _cfg.CM_Enumerate_Classes, (DWORD, ULONG, PGUID, ULONG)
    )

def CM_Enumerate_Classes(idx, flags=0):
    guid = GUID()
    raise_on_cr(_CM_Enumerate_Classes(idx, ref(guid), flags))
    return guid

################################################################################

_CM_Get_Parent = fun_fact(
    _cfg.CM_Get_Parent, (DWORD, PDWORD, DWORD, ULONG)
    )

def CM_Get_Parent(devinst):
    parent = DWORD()
    raise_on_cr(_CM_Get_Parent(ref(parent), devinst, 0))
    return parent.value

################################################################################

_CM_Request_Device_Eject = fun_fact(
    _cfg.CM_Request_Device_EjectW, (
        DWORD,
        DWORD,
        PINT,
        PWSTR,
        ULONG,
        ULONG
        )
    )

def CM_Request_Device_Eject(devinst):
    veto_type = INT()
    veto_name = string_buffer(MAX_PATH)
    err = _CM_Request_Device_Eject(
        devinst,
        ref(veto_type),
        veto_name,
        MAX_PATH,
        0
        )
    if err != CR_SUCCESS or veto_type.value != 0:
        vv = veto_type.value
        vn = veto_name.value
        def_err = ERROR_CURRENT_DIRECTORY
        err = CM_MapCrToWin32Err(err, def_err) or def_err
        raise OSError(err, f"device removal was vetoed ({vv}): {vn}")

################################################################################

_CM_Locate_DevNode = fun_fact(
    _cfg.CM_Locate_DevNodeW, (DWORD, PDWORD, PWSTR, ULONG)
    )

def CM_Locate_DevNode(device_id, flags=CM_LOCATE_DEVNODE_NORMAL):
    devinst = DWORD()
    raise_on_cr(_CM_Locate_DevNode(ref(devinst), device_id, flags))
    return devinst.value

################################################################################

_CM_Get_DevNode_Registry_Property = fun_fact(
    _cfg.CM_Get_DevNode_Registry_PropertyW, (
        DWORD,
        DWORD,
        ULONG,
        PULONG,
        PVOID,
        PULONG,
        ULONG
        )
    )

def CM_Get_DevNode_Registry_Property(devinst, prop):
    reg_type = DWORD()
    req_size = DWORD()
    _CM_Get_DevNode_Registry_Property(
        devinst,
        prop,
        ref(reg_type),
        None,
        ref(req_size),
        0
        )
    buf = byte_buffer(req_size.value)
    raise_on_cr(
        _CM_Get_DevNode_Registry_Property(
            devinst,
            prop,
            ref(reg_type),
            buf,
            ref(req_size),
            0
            )
        )
    return registry_to_py(reg_type.value, buf.raw[:req_size.value])

################################################################################

_CM_Get_Device_Interface_List_Size = fun_fact(
    _cfg.CM_Get_Device_Interface_List_SizeW, (
        DWORD,
        PULONG,
        PGUID,
        PWSTR,
        ULONG
        )
    )

def CM_Get_Device_Interface_List_Size(
    guid,
    didstr,
    flags=CM_GET_DEVICE_INTERFACE_LIST_PRESENT
    ):
    size = ULONG()
    raise_on_cr(
        _CM_Get_Device_Interface_List_Size(
            ref(size),
            ref(guid),
            didstr,
            flags
            )
        )
    return size.value

################################################################################

_CM_Get_Device_Interface_List = fun_fact(
    _cfg.CM_Get_Device_Interface_ListW, (
        DWORD,
        PGUID,
        PWSTR,
        PWSTR,
        ULONG,
        ULONG
        )
    )

def CM_Get_Device_Interface_List(
    guid,
    didstr,
    flags=CM_GET_DEVICE_INTERFACE_LIST_PRESENT
    ):
    res = CR_BUFFER_SMALL
    while res == CR_BUFFER_SMALL:
        size = ULONG(CM_Get_Device_Interface_List_Size(guid, didstr, flags))
        iface = string_buffer(size.value)
        res = _CM_Get_Device_Interface_List(
            ref(guid),
            didstr,
            iface,
            size,
            flags
            )
        if res == CR_SUCCESS:
            return multi_str_from_str(iface.value)
        if res != CR_BUFFER_SMALL:
            raise_on_cr(res)
    return []

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from ctwin32 import advapi
import datetime

BOOT_EVENT_ID = 6009

def boot_time():
    for e in advapi.enum_event_log("System"):
        if (e.EventID & 0xffff) == BOOT_EVENT_ID:
            return e.TimeGenerated
    raise OSError("no boot event found")

def up_time(time_boot=None):
    if time_boot is None:
        time_boot = boot_time()
    utime = datetime.datetime.now(time_boot.tzinfo) - time_boot
    # ignore milliseconds
    return datetime.timedelta(seconds=int(utime.total_seconds()))

if __name__ == "__main__":
    t_boot = boot_time()
    print(f"This computer was booted on {t_boot}.")
    print(f"It has been running for {up_time(t_boot)}.")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    BOOL,
    BYTE,
    byte_buffer,
    DWORD,
    HANDLE,
    INT,
    LONG,
    PINT,
    PLOGFONT,
    POINTER,
    PRECT,
    PVOID,
    PWSTR,
    UINT,
    WCHAR,
    )
from . import (
    ref,
    raise_if,
    raise_on_zero,
    fun_fact,
    ETO_OPAQUE,
    HGDI_ERROR,
    )

_gdi = ctypes.WinDLL("gdi32.dll", use_last_error=True)

################################################################################

GetDeviceCaps = fun_fact(_gdi.GetDeviceCaps, (INT, HANDLE, INT))

################################################################################

_CreateFontIndirect = fun_fact(_gdi.CreateFontIndirectW, (HANDLE, PLOGFONT))

def CreateFontIndirect(lf):
    res = _CreateFontIndirect(ref(lf))
    raise_on_zero(res)
    return res

################################################################################

_SelectObject = fun_fact(_gdi.SelectObject, (HANDLE, HANDLE, HANDLE))

def SelectObject(hdc, hobj):
    hprev = _SelectObject(hdc, hobj)
    raise_if(not hprev or hprev == HGDI_ERROR)
    return hprev

################################################################################

_DeleteObject = fun_fact(_gdi.DeleteObject, (BOOL, HANDLE))

def DeleteObject(hobj):
    raise_on_zero(_DeleteObject(hobj))

################################################################################

class TEXTMETRIC(ctypes.Structure):
    _fields_ = (
        ("tmHeight", LONG),
        ("tmAscent", LONG),
        ("tmDescent", LONG),
        ("tmInternalLeading", LONG),
        ("tmExternalLeading", LONG),
        ("tmAveCharWidth", LONG),
        ("tmMaxCharWidth", LONG),
        ("tmWeight", LONG),
        ("tmOverhang", LONG),
        ("tmDigitizedAspectX", LONG),
        ("tmDigitizedAspectY", LONG),
        ("tmFirstChar", WCHAR),
        ("tmLastChar", WCHAR),
        ("tmDefaultChar", WCHAR),
        ("tmBreakChar", WCHAR),
        ("tmItalic", BYTE),
        ("tmUnderlined", BYTE),
        ("tmStruckOut", BYTE),
        ("tmPitchAndFamily", BYTE),
        ("tmCharSet", BYTE),
        )
PTEXTMETRIC = POINTER(TEXTMETRIC)

################################################################################

_GetTextMetrics = fun_fact(_gdi.GetTextMetricsW, (BOOL, HANDLE, PTEXTMETRIC))

def GetTextMetrics(hdc):
    tm = TEXTMETRIC()
    raise_on_zero(_GetTextMetrics(hdc, ref(tm)))
    return tm

################################################################################

_GetStockObject = fun_fact(_gdi.GetStockObject, (HANDLE, INT))

def GetStockObject(idx):
    obj = _GetStockObject(idx)
    raise_on_zero(obj)
    return obj

################################################################################

_SetBkMode = fun_fact(_gdi.SetBkMode, (INT, HANDLE, INT))

def SetBkMode(hdc, mode):
    previous = _SetBkMode(hdc, mode)
    raise_on_zero(previous)
    return previous

################################################################################

_SetBkColor = fun_fact(_gdi.SetBkColor, (DWORD, HANDLE, DWORD))

def SetBkColor(hdc, colorref):
    previous = _SetBkColor(hdc, colorref)
    raise_on_zero(previous)
    return previous

################################################################################

_TextOut = fun_fact(_gdi.TextOutW, (BOOL, HANDLE, INT, INT, PWSTR, INT))

def TextOut(hdc, x, y, text):
    raise_on_zero(_TextOut(hdc, x, y, text, len(text)))

################################################################################

_ExtTextOut = fun_fact(
    _gdi.ExtTextOutW,
    (BOOL, HANDLE, INT, INT, UINT, PRECT, PWSTR, UINT, PINT)
    )

def ExtTextOut(hdc, x, y, opt, rect, text, pDX=None):
    span = len(text) if text else 0
    raise_on_zero(
        _ExtTextOut(hdc, x, y, opt, ref(rect), text, span, pDX)
        )

################################################################################

def fill_solid_rect(hdc, rect, colorref):
    oldclr = SetBkColor(hdc, colorref)
    ExtTextOut(hdc, 0, 0, ETO_OPAQUE, rect, None, None)
    SetBkColor(hdc, oldclr)

################################################################################

_GetObject = fun_fact(_gdi.GetObjectW, (INT, HANDLE, INT, PVOID))

def GetObject(hdl):
    raise_on_zero(size := _GetObject(hdl, 0, None))
    buf = byte_buffer(size)
    raise_on_zero(_GetObject(hdl, size, buf))
    return buf

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    byte_buffer,
    BOOLEAN,
    ENDIANNESS,
    LONG,
    PVOID,
    ULONG,
    )
from .ntdll import raise_failed_status
from . import (
    ref,
    fun_fact,
    raise_on_zero,
    ns_from_struct,
    SystemExecutionState,
    ProcessorInformation,
    kernel,
    )

################################################################################

_popro = ctypes.WinDLL("powrprof.dll", use_last_error=True)

_CallNtPowerInformation = fun_fact(
    _popro.CallNtPowerInformation,
    (LONG, LONG, PVOID, ULONG, PVOID, ULONG),
    )

def CallNtPowerInformation(level, outsize=0, input=None):
    src, slen = (None, 0) if not input else (ref(input), ULONG(len(input)))
    dst, dlen = byte_buffer(outsize), ULONG(outsize)
    raise_failed_status(_CallNtPowerInformation(level, src, slen, dst, dlen))
    return dst.raw

################################################################################

_PowerInformationWithPrivileges = fun_fact(
    _popro.PowerInformationWithPrivileges,
    (LONG, LONG, PVOID, ULONG, PVOID, ULONG),
    )

def PowerInformationWithPrivileges(level, outsize=0, input=None):
    src, slen = (None, 0) if not input else (ref(input), ULONG(len(input)))
    dst, dlen = byte_buffer(outsize), ULONG(outsize)
    raise_failed_status(
        _PowerInformationWithPrivileges(level, src, slen, dst, dlen)
        )
    return dst

################################################################################

def get_system_execution_state():
    bts = CallNtPowerInformation(SystemExecutionState, ctypes.sizeof(ULONG))
    # result is a combination of ES_SYSTEM_REQUIRED, ES_DISPLAY_REQUIRED,
    # ES_USER_PRESENT, ES_AWAYMODE_REQUIRED and ES_CONTINUOUS
    return int.from_bytes(bts, byteorder=ENDIANNESS, signed=False)

################################################################################

class PROCESSOR_POWER_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("Number", ULONG),
        ("MaxMhz", ULONG),
        ("CurrentMhz", ULONG),
        ("MhzLimit", ULONG),
        ("MaxIdleState", ULONG),
        ("CurrentIdleState", ULONG),
        )

def get_system_processor_power_info():
    nump = kernel.GetSystemInfo().dwNumberOfProcessors
    PPIN = PROCESSOR_POWER_INFORMATION * nump
    bts = CallNtPowerInformation(ProcessorInformation, ctypes.sizeof(PPIN))
    return [ns_from_struct(i) for i in PPIN.from_buffer_copy(bts)]

################################################################################

_SetSuspendState = fun_fact(
    _popro.SetSuspendState,
    (BOOLEAN, BOOLEAN, BOOLEAN, BOOLEAN)
    )

def SetSuspendState(hibernate, force, wakeup_events_disabled):
    raise_on_zero(_SetSuspendState(hibernate, force, wakeup_events_disabled))

################################################################################




################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import time
import uuid
from pathlib import Path
from importlib.util import spec_from_file_location, module_from_spec

# Note: Since this file will not only serve as a module of its parent package
#       but also as a standalone script, we cannot use 'from . import ('.
from ctwin32 import (
    advapi,
    cmdline_from_args,
    kernel,
    wtypes,
    CREATE_NEW_CONSOLE,
    DELETE,
    ERROR_SERVICE_EXISTS,
    MAXIMUM_ALLOWED,
    NORMAL_PRIORITY_CLASS,
    PROCESS_QUERY_LIMITED_INFORMATION,
    SC_MANAGER_CONNECT,
    SC_MANAGER_CREATE_SERVICE,
    SecurityIdentification,
    SERVICE_ACCEPT_STOP,
    SERVICE_DEMAND_START,
    SERVICE_ERROR_NORMAL,
    SERVICE_QUERY_STATUS,
    SERVICE_RUNNING,
    SERVICE_START,
    SERVICE_STOPPED,
    SERVICE_WIN32_OWN_PROCESS,
    TokenPrimary,
    TokenSessionId,
    TOKEN_DUPLICATE,
    )

################################################################################

_FILE = Path(__file__).resolve()

################################################################################

def _load_func(file_name, func_name):
    dont_write_bytecode = sys.dont_write_bytecode
    sys.dont_write_bytecode = True

    file_path = Path(file_name)
    module_name = file_path.stem
    if module_name in sys.modules:
        raise ImportError(f"already a module: {module_name}")
    spec = spec_from_file_location(module_name, file_path)
    module = module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    sys.dont_write_bytecode = dont_write_bytecode
    return getattr(module, func_name)

################################################################################

@advapi.SERVICE_MAIN_FUNCTION
def _service_main(argc, argv):
    # Since this is a python callback that ctypes calls when requested
    # by foreign C code, ctypes has no way of propagating any exception
    # back to the python interpreter - those would simply be ignored.
    # Therefore any exception has to terminate the process.
    with kernel.terminate_on_exception():
        tid = kernel.GetCurrentThreadId()
        kernel.dbg_print(f"in service_main: {tid}")

        # convert PPWSTR to list of strings
        arglist = [argv[i] for i in range(argc)]

        # arguments for service_main will be:
        #   0       service name
        #   1       path to module implementing the function to be run
        #   2       name of the function to be run
        #   3 ...   string arguments for function

        for i, arg in enumerate(arglist):
            kernel.dbg_print(f"service arg[{i}]: {arg}")

        @advapi.HANDLER_FUNCTION
        def handler(control):
            # Since this service is expected to be very short-lived, we simply
            # ignore control requests.
            pass

        hdl = advapi.RegisterServiceCtrlHandler(arglist[0], handler)
        svc_stat = advapi.SERVICE_STATUS(
            SERVICE_WIN32_OWN_PROCESS,
            SERVICE_RUNNING,
            SERVICE_ACCEPT_STOP
            )
        advapi.SetServiceStatus(hdl, svc_stat)

        func = _load_func(arglist[1], arglist[2])
        fargs = arglist[3:]
        str_args = f"*{fargs}" if fargs else ""
        kernel.dbg_print(f"calling '{arglist[2]}({str_args})'")
        func(*fargs)

        svc_stat.dwCurrentState = SERVICE_STOPPED
        advapi.SetServiceStatus(hdl, svc_stat)
        kernel.dbg_print(f"returning from _service main: {tid}")

################################################################################

def _run_service():
    tid = kernel.GetCurrentThreadId()
    kernel.dbg_print(f"trying to start dispatcher: {tid}")

    # For a service with SERVICE_WIN32_OWN_PROCESS set, windows will ignore
    # the service name in the dispatch table.
    table = (advapi.SERVICE_TABLE_ENTRY * 2)()
    table[0].lpServiceName = ""
    table[0].lpServiceProc = _service_main
    # table[1] will be zero initialized by default, which is exactly
    # what we need

    # Calling StartServiceCtrlDispatcher will fail if this process was not
    # started by the SCM and is not supposed to run as a service.
    advapi.StartServiceCtrlDispatcher(table)
    kernel.dbg_print(f"returned from dispatcher: {tid}")

################################################################################

def _start_as_service(arglist):
    # Arguments for _start_as_service will be:
    #   0       path to module implementing the function to be run
    #   1       name of the function to be run
    #   2 ...   string arguments for function
    #
    # But _start_as_service doesn't really care. However, _service_main will
    # very well take this into account.

    service_name = str(uuid.uuid4())
    with advapi.OpenSCManager(None, None, SC_MANAGER_CREATE_SERVICE) as scm:
        create_args = (
            scm,
            service_name,
            service_name,
            SERVICE_START | SERVICE_QUERY_STATUS | DELETE,
            SERVICE_WIN32_OWN_PROCESS,
            SERVICE_DEMAND_START,
            SERVICE_ERROR_NORMAL,
            cmdline_from_args([sys.executable, _FILE, "_run_service"])
            )
        for _ in range(2):
            try:
                with advapi.CreateService(*create_args) as svc:
                    advapi.StartService(svc, arglist)
                    advapi.DeleteService(svc)
                break
            except OSError as e:
                kernel.dbg_print(f"create/start err: {e}")
                if e.winerror == ERROR_SERVICE_EXISTS:
                    with advapi.OpenService(scm, service_name, DELETE) as svc:
                        advapi.DeleteService(svc)
                else:
                    raise

################################################################################

def _prepare_service_args(file_name, func_name, arglist):
    if arglist is None:
        arglist = []
    service_args = []
    for arg in [file_name, func_name, *arglist]:
        if isinstance(arg, int | Path):
            service_args.append(str(arg))
        elif not isinstance(arg, str):
            raise TypeError("need str, got '{arg}' ({type(arg).__name__})")
        else:
            service_args.append(arg)

    # ensure import will not raise an exception
    func = _load_func(service_args[0], service_args[1])
    if not callable(func):
        raise TypeError(f"'{func_name}' must be callable")

    return service_args

################################################################################

def func_as_system(file_name, func_name, arglist=None):
    # `func_name` must be defined in the toplevel namespace of `file_name`
    service_args = _prepare_service_args(file_name, func_name, arglist)
    _start_as_service(service_args)

################################################################################

def create_process_in_session_copy_token(session, pid, arglist, flags=None):
    pqli = PROCESS_QUERY_LIMITED_INFORMATION
    with kernel.OpenProcess(pqli, False, pid) as hproc:
        with advapi.OpenProcessToken(hproc, TOKEN_DUPLICATE) as htok:
            # Cannot set the session ID of a token that is in use by a process.
            # Therefore we have to duplicate the token.
            duplicate_token_args = (
                htok,
                MAXIMUM_ALLOWED,
                kernel.SECURITY_ATTRIBUTES(),
                SecurityIdentification,
                TokenPrimary,
                )
            with advapi.DuplicateTokenEx(*duplicate_token_args) as hdup:
                # set the session
                session = wtypes.ULONG(session)
                advapi.SetTokenInformation(hdup, TokenSessionId, session)

                si = kernel.STARTUPINFO()
                si.lpDesktop = r"WinSta0\default"
                if flags is None:
                    flags = NORMAL_PRIORITY_CLASS | CREATE_NEW_CONSOLE
                with advapi.create_process_as_user(hdup, arglist, flags, si):
                    pass

################################################################################

def _call_func_as_svc_acc(*args):
    # args[0] : name of file defining function name to be called
    # args[1] : function name to be called
    # args[2:]: arguments for function
    kernel.dbg_print("in _call_func_as_svc_acc")
    for i, arg in enumerate(args):
        kernel.dbg_print(f"arg[{i}]: {arg}")

    func = _load_func(args[0], args[1])
    fargs = args[2:]
    str_args = f"*{fargs}" if fargs else ""
    kernel.dbg_print(f"calling '{args[1]}({str_args})'")
    func(*fargs)

################################################################################

def get_service_pid(name):
    with advapi.OpenSCManager(None, None, SC_MANAGER_CONNECT) as scm:
        flags = SERVICE_START | SERVICE_QUERY_STATUS
        with advapi.OpenService(scm, name, flags) as svc:
            status = advapi.QueryServiceStatusEx(svc)
            if status.dwProcessId:
                return status.dwProcessId
            else:
                advapi.StartService(svc, [])
                deadline = time.time() + 2
                status = advapi.QueryServiceStatusEx(svc)
                while not status.dwProcessId:
                    if time.time() >= deadline:
                        raise OSError("timeout waiting for service to start")
                    status = advapi.QueryServiceStatusEx(svc)
                return status.dwProcessId

################################################################################

def _service_account_from_system(*args):
    kernel.dbg_print("in _service_account_from_system")
    for i, arg in enumerate(args):
        kernel.dbg_print(f"arg[{i}]: {arg}")

    # possible arg structures
    # a) service_name cwd session file_name func_name *func_args
    # b) service_name 'command' cwd session *command_line_args
    pid = get_service_pid(args[0])
    if args[1] == "command":
        directory = args[2]
        session = int(args[3])
        arglist = args[4:]
    else:
        directory = args[1]
        session = int(args[2])
        arglist = [sys.executable, str(_FILE), "_call_func_as_svc_acc"]
        arglist.extend(args[3:])

    kernel.SetCurrentDirectory(directory)
    create_process_in_session_copy_token(session, pid, arglist)
    kernel.dbg_print("process was created")

################################################################################

def func_as_trusted_installer(file_name, func_name, arglist=None):
    # `func_name` must be defined in the toplevel namespace of `file_name`
    arglist = _prepare_service_args(file_name, func_name, arglist)
    ti_args = [
        "TrustedInstaller",
        Path.cwd(),
        kernel.ProcessIdToSessionId(kernel.GetCurrentProcessId()),
        *arglist
        ]
    ti_args = _prepare_service_args(
        _FILE,
        _service_account_from_system.__name__,
        ti_args
        )
    _start_as_service(ti_args)

################################################################################

def proc_as_trusted_installer(*command_line_args):
    arglist = [
        "TrustedInstaller",
        "command",
        Path.cwd(),
        kernel.ProcessIdToSessionId(kernel.GetCurrentProcessId()),
        *command_line_args
        ]
    _start_as_service(
        _prepare_service_args(
            _FILE,
            _service_account_from_system.__name__,
            arglist
            )
        )

################################################################################

def running_as_trusted_installer():
    ti_sid = "S-1-5-80-956008885-3418522649-1831038044-1853292631-2271478464"
    return advapi.CheckTokenMembership(
        None,
        advapi.ConvertStringSidToSid(ti_sid)
        )

################################################################################

def _main():
    kernel.dbg_print("svc_util in main")
    with kernel.terminate_on_exception():
        kernel.dbg_print(f"starting {_FILE.name}: {sys.argv}")
        available_tasks = (
            # argv[1]/func          min len(argv)
            (_run_service,          2),
            (_call_func_as_svc_acc, 4),
            )
        for func, length in available_tasks:
            if len(sys.argv) >= length and sys.argv[1] == func.__name__:
                func(*sys.argv[2:])
                break
        else:
            raise RuntimeError(f"invalid args: {sys.argv}")

################################################################################

if __name__ == "__main__":
    _main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from ctwin32 import shell, wtypes

E_FILE_NOT_FOUND = wtypes.LONG(0x80070002).value

for name in shell.__dict__:
    if name.startswith("CSIDL_"):
        csidl = getattr(shell, name)
        if (csidl & shell.CSIDL_FLAG_MASK) == 0:
            try:
                folder = shell.SHGetFolderPath(csidl)
            except OSError as e:
                if e.winerror == E_FILE_NOT_FOUND:
                    nvid = csidl | shell.CSIDL_FLAG_DONT_VERIFY
                    try:
                        folder = "* " + shell.SHGetFolderPath(nvid)
                    except OSError as e:
                        folder = str(e)
                else:
                    folder = str(e)
            print(f"{name:30} {csidl:4x} = {folder}")

print("\n* : non-existent")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This example shows how to wait for a job object. I.e. waiting until all
# processes in this job stopped executing.

import sys
from ctwin32 import (
    kernel,
    CREATE_SUSPENDED,
    INFINITE,
    INVALID_HANDLE_VALUE,
    JobObjectAssociateCompletionPortInformation,
    JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO
    )

################################################################################

def create_job(arglist):
    arglist = ["cmd", "/c", *arglist] if sys.stdout is not None else arglist
    job = kernel.CreateJobObject()
    with kernel.create_process(arglist, CREATE_SUSPENDED) as proc_info:
        kernel.AssignProcessToJobObject(job, proc_info.hProcess)
        io_port = kernel.create_io_completion_port(INVALID_HANDLE_VALUE, 0)
        joacp = kernel.JOBOBJECT_ASSOCIATE_COMPLETION_PORT(job, io_port)
        kernel.SetInformationJobObject(
            job,
            JobObjectAssociateCompletionPortInformation,
            joacp
            )
        kernel.ResumeThread(proc_info.hThread)
    return job, io_port

################################################################################

def job_is_still_alive(job, io_port, timeout=INFINITE):
    code, key, _ = kernel.GetQueuedCompletionStatus(io_port, timeout)
    return key != job or code != JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO

################################################################################

if __name__ == "__main__":
    job, io_port = create_job(sys.argv[1:])
    while (job_is_still_alive(job.value, io_port)):
        pass

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import string_buffer, DWORD, PWSTR
from . import fun_fact, raise_on_zero

################################################################################

_dbghlp = ctypes.WinDLL("dbghelp.dll", use_last_error=True)

UNDNAME_COMPLETE = 0x0000
UNDNAME_NO_LEADING_UNDERSCORES = 0x0001
UNDNAME_NO_MS_KEYWORDS = 0x0002
UNDNAME_NO_FUNCTION_RETURNS = 0x0004
UNDNAME_NO_ALLOCATION_MODEL = 0x0008
UNDNAME_NO_ALLOCATION_LANGUAGE = 0x0010
UNDNAME_NO_MS_THISTYPE = 0x0020
UNDNAME_NO_CV_THISTYPE = 0x0040
UNDNAME_NO_THISTYPE = 0x0060
UNDNAME_NO_ACCESS_SPECIFIERS = 0x0080
UNDNAME_NO_THROW_SIGNATURES = 0x0100
UNDNAME_NO_MEMBER_TYPE = 0x0200
UNDNAME_NO_RETURN_UDT_MODEL = 0x0400
UNDNAME_32_BIT_DECODE = 0x0800
UNDNAME_NAME_ONLY = 0x1000
UNDNAME_NO_ARGUMENTS = 0x2000
UNDNAME_NO_SPECIAL_SYMS = 0x4000

################################################################################

_UnDecorateSymbolName = fun_fact(
    _dbghlp.UnDecorateSymbolNameW, (DWORD, PWSTR, PWSTR, DWORD, DWORD)
    )

def UnDecorateSymbolName(sym_name, flags=UNDNAME_COMPLETE):
    size = 256
    while True:
        buf = string_buffer(size)
        result = _UnDecorateSymbolName(sym_name, buf, size, flags)
        raise_on_zero(result)
        if result <= size - 4:
            return buf.value
        size *= 2

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    ENDIANNESS,
    HANDLE,
    NTSTATUS,
    PCHAR,
    PHANDLE,
    PULONG,
    PVOID,
    PWSTR,
    ScdToBeClosed,
    ULONG,
    )
from . import ref, fun_fact
from .ntdll import raise_failed_status, STATUS_INVALID_SIGNATURE

_bcr = ctypes.WinDLL("bcrypt.dll", use_last_error=True)

################################################################################

BCRYPT_OBJECT_LENGTH        = "ObjectLength"
BCRYPT_ALGORITHM_NAME       = "AlgorithmName"
BCRYPT_PROVIDER_HANDLE      = "ProviderHandle"
BCRYPT_CHAINING_MODE        = "ChainingMode"
BCRYPT_BLOCK_LENGTH         = "BlockLength"
BCRYPT_KEY_LENGTH           = "KeyLength"
BCRYPT_KEY_OBJECT_LENGTH    = "KeyObjectLength"
BCRYPT_KEY_STRENGTH         = "KeyStrength"
BCRYPT_KEY_LENGTHS          = "KeyLengths"
BCRYPT_BLOCK_SIZE_LIST      = "BlockSizeList"
BCRYPT_EFFECTIVE_KEY_LENGTH = "EffectiveKeyLength"
BCRYPT_HASH_LENGTH          = "HashDigestLength"
BCRYPT_HASH_OID_LIST        = "HashOIDList"
BCRYPT_PADDING_SCHEMES      = "PaddingSchemes"
BCRYPT_SIGNATURE_LENGTH     = "SignatureLength"
BCRYPT_HASH_BLOCK_LENGTH    = "HashBlockLength"
BCRYPT_AUTH_TAG_LENGTH      = "AuthTagLength"
BCRYPT_PRIMITIVE_TYPE       = "PrimitiveType"
BCRYPT_IS_KEYED_HASH        = "IsKeyedHash"
BCRYPT_IS_REUSABLE_HASH     = "IsReusableHash"
BCRYPT_MESSAGE_BLOCK_LENGTH = "MessageBlockLength"
BCRYPT_PUBLIC_KEY_LENGTH    = "PublicKeyLength"

BCRYPT_INITIALIZATION_VECTOR = "IV"

BCRYPT_CHAIN_MODE_NA  = "ChainingModeN/A"
BCRYPT_CHAIN_MODE_CBC = "ChainingModeCBC"
BCRYPT_CHAIN_MODE_ECB = "ChainingModeECB"
BCRYPT_CHAIN_MODE_CFB = "ChainingModeCFB"
BCRYPT_CHAIN_MODE_CCM = "ChainingModeCCM"
BCRYPT_CHAIN_MODE_GCM = "ChainingModeGCM"

BCRYPT_PUBLIC_KEY_BLOB     = "PUBLICBLOB"
BCRYPT_PRIVATE_KEY_BLOB    = "PRIVATEBLOB"
BCRYPT_ECCPUBLIC_BLOB      = "ECCPUBLICBLOB"
BCRYPT_ECCPRIVATE_BLOB     = "ECCPRIVATEBLOB"
BCRYPT_RSAPUBLIC_BLOB      = "RSAPUBLICBLOB"
BCRYPT_RSAPRIVATE_BLOB     = "RSAPRIVATEBLOB"
LEGACY_RSAPUBLIC_BLOB      = "CAPIPUBLICBLOB"
LEGACY_RSAPRIVATE_BLOB     = "CAPIPRIVATEBLOB"
BCRYPT_ECCFULLPUBLIC_BLOB  = "ECCFULLPUBLICBLOB"
BCRYPT_ECCFULLPRIVATE_BLOB = "ECCFULLPRIVATEBLOB"
SSL_ECCPUBLIC_BLOB         = "SSLECCPUBLICBLOB"
BCRYPT_OPAQUE_KEY_BLOB     = "OpaqueKeyBlob"
BCRYPT_KEY_DATA_BLOB       = "KeyDataBlob"
BCRYPT_AES_WRAP_KEY_BLOB   = "Rfc3565KeyWrapBlob"

BCRYPT_RSA_ALGORITHM               = "RSA"
BCRYPT_RSA_SIGN_ALGORITHM          = "RSA_SIGN"
BCRYPT_DH_ALGORITHM                = "DH"
BCRYPT_DSA_ALGORITHM               = "DSA"
BCRYPT_RC2_ALGORITHM               = "RC2"
BCRYPT_RC4_ALGORITHM               = "RC4"
BCRYPT_AES_ALGORITHM               = "AES"
BCRYPT_DES_ALGORITHM               = "DES"
BCRYPT_DESX_ALGORITHM              = "DESX"
BCRYPT_3DES_ALGORITHM              = "3DES"
BCRYPT_3DES_112_ALGORITHM          = "3DES_112"
BCRYPT_MD2_ALGORITHM               = "MD2"
BCRYPT_MD4_ALGORITHM               = "MD4"
BCRYPT_MD5_ALGORITHM               = "MD5"
BCRYPT_SHA1_ALGORITHM              = "SHA1"
BCRYPT_SHA256_ALGORITHM            = "SHA256"
BCRYPT_SHA384_ALGORITHM            = "SHA384"
BCRYPT_SHA512_ALGORITHM            = "SHA512"
BCRYPT_AES_GMAC_ALGORITHM          = "AES-GMAC"
BCRYPT_AES_CMAC_ALGORITHM          = "AES-CMAC"
BCRYPT_ECDSA_P256_ALGORITHM        = "ECDSA_P256"
BCRYPT_ECDSA_P384_ALGORITHM        = "ECDSA_P384"
BCRYPT_ECDSA_P521_ALGORITHM        = "ECDSA_P521"
BCRYPT_ECDH_P256_ALGORITHM         = "ECDH_P256"
BCRYPT_ECDH_P384_ALGORITHM         = "ECDH_P384"
BCRYPT_ECDH_P521_ALGORITHM         = "ECDH_P521"
BCRYPT_RNG_ALGORITHM               = "RNG"
BCRYPT_RNG_FIPS186_DSA_ALGORITHM   = "FIPS186DSARNG"
BCRYPT_RNG_DUAL_EC_ALGORITHM       = "DUALECRNG"
BCRYPT_SP800108_CTR_HMAC_ALGORITHM = "SP800_108_CTR_HMAC"
BCRYPT_SP80056A_CONCAT_ALGORITHM   = "SP800_56A_CONCAT"
BCRYPT_PBKDF2_ALGORITHM            = "PBKDF2"
BCRYPT_CAPI_KDF_ALGORITHM          = "CAPI_KDF"
BCRYPT_TLS1_1_KDF_ALGORITHM        = "TLS1_1_KDF"
BCRYPT_TLS1_2_KDF_ALGORITHM        = "TLS1_2_KDF"
BCRYPT_ECDSA_ALGORITHM             = "ECDSA"
BCRYPT_ECDH_ALGORITHM              = "ECDH"
BCRYPT_XTS_AES_ALGORITHM           = "XTS-AES"
BCRYPT_HKDF_ALGORITHM              = "HKDF"

################################################################################

_BCryptDestroyKey = fun_fact(
    _bcr.BCryptDestroyKey, (NTSTATUS, HANDLE)
    )

def BCryptDestroyKey(key):
    raise_failed_status(_BCryptDestroyKey(key))

class BCRYPT_KEY(ScdToBeClosed, HANDLE, close_func=BCryptDestroyKey, invalid=0):
    def __init__(self, init=None):
        super().__init__(init)
        # additional buffer for symmetric keys
        self.buf = None

    def close(self):
        super().close()
        self.buf = None

################################################################################

_BCryptDestroyHash = fun_fact(
    _bcr.BCryptDestroyHash, (NTSTATUS, HANDLE)
    )

def BCryptDestroyHash(hash):
    raise_failed_status(_BCryptDestroyHash(hash))

class BCRYPT_HASH(
        ScdToBeClosed,
        HANDLE,
        close_func=BCryptDestroyHash,
        invalid=0
        ):
    pass

################################################################################

_BCryptCloseAlgorithmProvider = fun_fact(
    _bcr.BCryptCloseAlgorithmProvider, (NTSTATUS, HANDLE, ULONG)
    )

def BCryptCloseAlgorithmProvider(balg):
    raise_failed_status(_BCryptCloseAlgorithmProvider(balg, 0))

class BCRYPT_ALG(
        ScdToBeClosed,
        HANDLE,
        close_func=BCryptCloseAlgorithmProvider,
        invalid=0
        ):
    pass

################################################################################

_BCryptOpenAlgorithmProvider = fun_fact(
    _bcr.BCryptOpenAlgorithmProvider, (NTSTATUS, PHANDLE, PWSTR, PWSTR, ULONG)
    )

def BCryptOpenAlgorithmProvider(alg, flags=0, impl=None):
    balg = BCRYPT_ALG()
    raise_failed_status(
        _BCryptOpenAlgorithmProvider(ref(balg), alg, impl, flags)
        )
    return balg

################################################################################

_BCryptGetProperty = fun_fact(
    _bcr.BCryptGetProperty, (
        NTSTATUS,
        HANDLE,
        PWSTR,
        PCHAR,
        ULONG,
        PULONG,
        ULONG
        )
    )

def BCryptGetProperty(obj, name):
    size = ULONG()
    raise_failed_status(_BCryptGetProperty(obj, name, None, 0, ref(size), 0))
    buf = byte_buffer(size.value)
    raise_failed_status(_BCryptGetProperty(obj, name, buf, size, ref(size), 0))
    return bytes(buf)

def get_property_ulong(obj, name):
    _bytes = BCryptGetProperty(obj, name)
    if len(_bytes) != ctypes.sizeof(ULONG):
        raise ValueError(f"property size mismatch ({len(_bytes)})")
    return int.from_bytes(_bytes, ENDIANNESS, signed=False)

################################################################################

_BCryptSetProperty = fun_fact(
    _bcr.BCryptSetProperty, (
        NTSTATUS,
        HANDLE,
        PWSTR,
        PCHAR,
        ULONG,
        ULONG
        )
    )

def BCryptSetProperty(obj, name, value):
    if isinstance(value, int):
        buf = value.to_bytes(((value.bit_length() + 31) // 32) * 4, ENDIANNESS)
    elif isinstance(value, str):
        buf = bytes(string_buffer(value))
    else:
        buf = bytes(value)
    raise_failed_status(_BCryptSetProperty(obj, name, buf, len(buf), 0))

################################################################################

_BCryptCreateHash = fun_fact(
    _bcr.BCryptCreateHash, (
        NTSTATUS,
        HANDLE,
        PHANDLE,
        PCHAR,
        ULONG,
        PULONG,
        ULONG,
        ULONG
        )
    )

def BCryptCreateHash(balg, obj_buf, secret=None):
    sec_size = ULONG(0 if secret is None else len(secret))
    bhash = BCRYPT_HASH()
    raise_failed_status(
        _BCryptCreateHash(
            balg,
            ref(bhash),
            obj_buf,
            len(obj_buf),
            secret,
            sec_size,
            0
            )
        )
    return bhash

################################################################################

_BCryptHashData = fun_fact(
    _bcr.BCryptHashData, (NTSTATUS, HANDLE, PCHAR, ULONG, ULONG)
    )

def BCryptHashData(bhash, data):
    raise_failed_status(_BCryptHashData(bhash, data, len(data), 0))

################################################################################

_BCryptFinishHash = fun_fact(
    _bcr.BCryptFinishHash, (NTSTATUS, HANDLE, PCHAR, ULONG, ULONG)
    )

def BCryptFinishHash(bhash, dig_size):
    buf = byte_buffer(dig_size)
    raise_failed_status(_BCryptFinishHash(bhash, buf, len(buf), 0))
    return bytes(buf)

################################################################################

class BCryptHash:
    "standard python hash wrapper for a BCrypt hash"

    def __init__(self, alg):
        self.alg = alg
        self.dig_size = 0
        self.hash = None
        self.obj_buf = None
        self.balg = None
        try:
            self.balg = BCryptOpenAlgorithmProvider(alg)
            self.dig_size = get_property_ulong(self.balg, BCRYPT_HASH_LENGTH)
            obj_size = get_property_ulong(self.balg, BCRYPT_OBJECT_LENGTH)
            self.obj_buf = byte_buffer(obj_size)
            self.hash = BCryptCreateHash(self.balg, self.obj_buf)
        except OSError:
            self.close()
            raise

    def close(self):
        if self.hash is not None:
            self.hash.close()
            self.hash = None
        if self.balg is not None:
            self.balg.close()
            self.balg = None
        self.obj_buf = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def copy(self):
        raise NotImplementedError(
            f"cannot copy {self.__class__.__name__} objects"
            )

    def update(self, data):
        BCryptHashData(self.hash, data)

    def digest(self):
        dig = BCryptFinishHash(self.hash, self.dig_size)

        # can only ask for digest once
        self.close()

        return dig

    def hexdigest(self):
        return self.digest().hex()

    @property
    def digest_size(self):
        return self.dig_size

    @property
    def block_size(self):
        return 42

    @property
    def name(self):
        return self.alg

################################################################################

_BCryptExportKey = fun_fact(
    _bcr.BCryptExportKey, (
        NTSTATUS,
        HANDLE,
        HANDLE,
        PWSTR,
        PCHAR,
        ULONG,
        PULONG,
        ULONG
        )
    )

def BCryptExportKey(key, btype, exp_key=None):
    size = ULONG()
    raise_failed_status(
        _BCryptExportKey(key, exp_key, btype, None, 0, ref(size), 0)
        )
    blob = byte_buffer(size.value)
    raise_failed_status(
        _BCryptExportKey(key, exp_key, btype, blob, size, ref(size), 0)
        )
    return bytes(blob)

################################################################################

_BCryptGenerateKeyPair = fun_fact(
    _bcr.BCryptGenerateKeyPair, (NTSTATUS, HANDLE, PHANDLE, ULONG, ULONG)
    )

def BCryptGenerateKeyPair(balg, key_len):
    key = BCRYPT_KEY()
    raise_failed_status(_BCryptGenerateKeyPair(balg, ref(key), key_len, 0))
    return key

################################################################################

_BCryptFinalizeKeyPair = fun_fact(
    _bcr.BCryptFinalizeKeyPair, (NTSTATUS, HANDLE, ULONG)
    )

def BCryptFinalizeKeyPair(key):
    raise_failed_status(_BCryptFinalizeKeyPair(key, 0))

################################################################################

_BCryptGenerateSymmetricKey = fun_fact(
    _bcr.BCryptGenerateSymmetricKey, (
        NTSTATUS,
        HANDLE,
        PHANDLE,
        PCHAR,
        ULONG,
        PCHAR,
        ULONG,
        ULONG
        )
    )

def BCryptGenerateSymmetricKey(balg, secret):
    ksize = get_property_ulong(balg, BCRYPT_OBJECT_LENGTH)
    key = BCRYPT_KEY()
    key.buf = byte_buffer(ksize)
    raise_failed_status(
        _BCryptGenerateSymmetricKey(
            balg,
            ref(key),
            key.buf,
            len(key.buf),
            secret,
            len(secret),
            0
            )
        )
    return key

################################################################################

_BCryptImportKeyPair = fun_fact(
    _bcr.BCryptImportKeyPair, (
        NTSTATUS,
        HANDLE,
        HANDLE,
        PWSTR,
        PHANDLE,
        PCHAR,
        ULONG,
        ULONG
        )
    )

def BCryptImportKeyPair(balg, btype, kbuf):
    key = BCRYPT_KEY()
    raise_failed_status(
        _BCryptImportKeyPair(balg, None, btype, ref(key), kbuf, len(kbuf), 0)
        )
    return key

################################################################################

_BCryptSignHash = fun_fact(
    _bcr.BCryptSignHash, (
        NTSTATUS,
        HANDLE,
        PVOID,
        PCHAR,
        ULONG,
        PCHAR,
        ULONG,
        PULONG,
        ULONG
        )
    )

def BCryptSignHash(key, digest, flags=0):
    size = ULONG()
    raise_failed_status(
        _BCryptSignHash(
            key,
            None,
            digest,
            len(digest),
            None,
            0,
            ref(size),
            flags
            )
        )
    signature = byte_buffer(size.value)
    raise_failed_status(
        _BCryptSignHash(
            key,
            None,
            digest,
            len(digest),
            signature,
            len(signature),
            ref(size),
            flags
            )
        )
    return bytes(signature)

################################################################################

_BCryptVerifySignature = fun_fact(
    _bcr.BCryptVerifySignature, (
        NTSTATUS,
        HANDLE,
        PVOID,
        PCHAR,
        ULONG,
        PCHAR,
        ULONG,
        ULONG
        )
    )

def BCryptVerifySignature(key, digest, signature):
    status = _BCryptVerifySignature(
        key,
        None,
        digest,
        len(digest),
        signature,
        len(signature),
        0
        )
    if status < 0 and status != STATUS_INVALID_SIGNATURE:
        raise_failed_status(status)
    return status == 0

################################################################################

_BCryptEncrypt = fun_fact(
    _bcr.BCryptEncrypt, (
        NTSTATUS,
        HANDLE,
        PCHAR,
        ULONG,
        PVOID,
        PCHAR,
        ULONG,
        PCHAR,
        ULONG,
        PULONG,
        ULONG
        )
    )

def BCryptEncrypt(key, input, iv=None, flags=0):
    size = ULONG()
    raise_failed_status(
        _BCryptEncrypt(
            key,
            input,
            len(input),
            None,
            iv,
            0 if iv is None else len(iv),
            None,
            0,
            ref(size),
            flags
            )
        )
    output = byte_buffer(size.value)
    raise_failed_status(
        _BCryptEncrypt(
            key,
            input,
            len(input),
            None,
            iv,
            0 if iv is None else len(iv),
            output,
            size,
            ref(size),
            flags
            )
        )
    return bytes(output)

################################################################################

_BCryptDecrypt = fun_fact(
    _bcr.BCryptDecrypt, (
        NTSTATUS,
        HANDLE,
        PCHAR,
        ULONG,
        PVOID,
        PCHAR,
        ULONG,
        PCHAR,
        ULONG,
        PULONG,
        ULONG
        )
    )

def BCryptDecrypt(key, input, iv=None, flags=0):
    size = ULONG()
    raise_failed_status(
        _BCryptDecrypt(
            key,
            input,
            len(input),
            None,
            iv,
            0 if iv is None else len(iv),
            None,
            0,
            ref(size),
            flags
            )
        )
    output = byte_buffer(size.value)
    raise_failed_status(
        _BCryptDecrypt(
            key,
            input,
            len(input),
            None,
            iv,
            0 if iv is None else len(iv),
            output,
            size,
            ref(size),
            flags
            )
        )
    return bytes(output)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This sample demonstrates how to open a console window where `cmd.exe` is
# running under the local system account (this requires administrative
# privileges). Hence the name: `local system console` -> `lsc`. If you examine
# this code more closely, you will find that it can do even more than that.
#
################################################################################

import sys
from pathlib import Path
from ctwin32 import kernel, advapi, svc_util

################################################################################

def to_be_run_as_system(*args):
    kernel.dbg_print(f"in to_be_run_as_system: {kernel.GetCurrentThreadId()}")

    # 1st arg: directory where to run
    # 2nd arg: session where to run
    #
    # possible arg structures
    # a) cwd session InitialCommandForCmdExe...
    # b) cwd session 'cpy_tok' ProcessId CommandLineToBeExecuted...

    if len(args) > 1:
        kernel.SetCurrentDirectory(args[0])
        session = int(args[1])
        if len(args) > 4 and args[2] == "cpy_tok":
            pid = int(args[3])
            arglist = args[4:]
        else:
            pid = kernel.GetCurrentProcessId()
            arglist = ["cmd.exe", "/K", "title", "LocalSystem", "&"]
            arglist.extend(args[2:])

        svc_util.create_process_in_session_copy_token(session, pid, arglist)
        kernel.dbg_print("process was created")

################################################################################

if __name__ == "__main__":
    if advapi.running_as_admin():
        session = kernel.ProcessIdToSessionId(kernel.GetCurrentProcessId())
        arglist = [Path.cwd(), session, *sys.argv[1:]]
        kernel.dbg_print("trying to run as system")
        svc_util.func_as_system(__file__, to_be_run_as_system.__name__, arglist)
    else:
        print("Running as system requires administrative privileges.")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import traceback
import datetime as _dt
import collections as _collections
from enum import IntEnum as _int_enum

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    BOOL,
    BYTE,
    CallbackContext,
    CallbackContextPtr,
    CHAR,
    DWORD,
    FILETIME,
    HANDLE,
    INT,
    LARGE_INTEGER,
    PBOOL,
    PBYTE,
    PDWORD,
    PLARGE_INTEGER,
    POINTER,
    PPWSTR,
    PSIZE_T,
    PULONG_PTR,
    PUSHORT,
    PVOID,
    PWIN32_FIND_DATA,
    PWSTR,
    ScdToBeClosed,
    SHORT,
    SIZE_T,
    SYSTEMTIME,
    UINT,
    UINT_PTR,
    ULONG,
    ULONGLONG,
    ULONG_PTR,
    USHORT,
    WCHAR,
    WCHAR_SIZE,
    WIN32_FIND_DATA,
    WinError,
    WORD,
    )
from . import (
    cmdline_from_args,
    fun_fact,
    multi_str_from_addr,
    multi_str_from_ubuf,
    ns_from_struct,
    ntdll,
    raise_if,
    raise_on_err,
    raise_on_zero,
    ref,
    ENABLE_VIRTUAL_TERMINAL_PROCESSING,
    ERROR_ACCESS_DENIED,
    ERROR_FILE_NOT_FOUND,
    ERROR_HANDLE_EOF,
    ERROR_INSUFFICIENT_BUFFER,
    ERROR_MORE_DATA,
    ERROR_NO_MORE_FILES,
    ERROR_PIPE_CONNECTED,
    ERROR_RESOURCE_ENUM_USER_STOP,
    ERROR_RESOURCE_NAME_NOT_FOUND,
    ERROR_SUCCESS,
    GENERIC_READ,
    GENERIC_WRITE,
    FILE_ATTRIBUTE_DIRECTORY,
    FILE_ATTRIBUTE_NORMAL,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    FILE_TYPE_CHAR,
    FILE_TYPE_UNKNOWN,
    INVALID_FILE_ATTRIBUTES,
    INVALID_HANDLE_VALUE,
    IMAGE_FILE_MACHINE_UNKNOWN,
    IMAGE_FILE_MACHINE_AMD64,
    IMAGE_FILE_MACHINE_I386,
    MB_OK,
    MB_ICONERROR,
    OPEN_EXISTING,
    PIPE_ACCESS_DUPLEX,
    PIPE_READMODE_MESSAGE,
    PIPE_REJECT_REMOTE_CLIENTS,
    PIPE_TYPE_MESSAGE,
    PIPE_UNLIMITED_INSTANCES,
    RT_MESSAGETABLE,
    STD_OUTPUT_HANDLE,
    THREAD_PRIORITY_ERROR_RETURN,
    WAIT_FAILED,
    )

_k32 = ctypes.WinDLL("kernel32.dll", use_last_error=True)

################################################################################

ExitProcess = fun_fact(_k32.ExitProcess, (None, UINT))

################################################################################

# Here we use the functions from ctypes instead of using GetLastError and
# SetLastError directly. We do so to work with the copies that ctypes manages.
# These are protected from being manipulated by the system calls of CPython
# itself. This requires that all DLLs involved are created with
# 'use_last_error=True'.
GetLastError = ctypes.get_last_error
SetLastError = ctypes.set_last_error

################################################################################

_LocalFree = fun_fact(_k32.LocalFree, (HANDLE, HANDLE))

def LocalFree(hmem):
    raise_if(_LocalFree(hmem))

################################################################################

_GlobalFree = fun_fact(_k32.GlobalFree, (HANDLE, HANDLE))

def GlobalFree(hmem):
    raise_if(_GlobalFree(hmem))

################################################################################

_GlobalAlloc = fun_fact(_k32.GlobalAlloc, (HANDLE, UINT, SIZE_T))

def GlobalAlloc(flags, size):
    res = _GlobalAlloc(flags, size)
    raise_on_zero(res)
    return res

################################################################################

_GlobalLock = fun_fact(_k32.GlobalLock, (PVOID, HANDLE))

def GlobalLock(hmem):
    res = _GlobalLock(hmem)
    raise_on_zero(res)
    return res

################################################################################

_GlobalUnlock = fun_fact(_k32.GlobalUnlock, (BOOL, HANDLE))

def GlobalUnlock(hmem):
    res = _GlobalUnlock(hmem)
    if not res and (err := GetLastError()) != ERROR_SUCCESS:
        raise WinError(err)

################################################################################

_CloseHandle = fun_fact(_k32.CloseHandle, (BOOL, HANDLE))

def CloseHandle(handle):
    raise_on_zero(_CloseHandle(handle))

################################################################################

class KHANDLE(ScdToBeClosed, HANDLE, close_func=CloseHandle, invalid=0):
    pass

PKHANDLE = POINTER(KHANDLE)

################################################################################

class FHANDLE(
        ScdToBeClosed,
        HANDLE,
        close_func=CloseHandle,
        invalid=INVALID_HANDLE_VALUE
        ):
    pass

PFHANDLE = POINTER(FHANDLE)

################################################################################

class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("nLength", DWORD),
        ("lpSecurityDescriptor", PVOID),
        ("bInheritHandle", BOOL),
        )

    def __init__(self):
        self.nLength = ctypes.sizeof(self)

PSECURITY_ATTRIBUTES = POINTER(SECURITY_ATTRIBUTES)

################################################################################

_CreateFile = fun_fact(
    _k32.CreateFileW, (
        HANDLE,
        PWSTR,
        DWORD,
        DWORD,
        PSECURITY_ATTRIBUTES,
        DWORD,
        DWORD,
        HANDLE
        )
    )

def CreateFile(file_name, access, share_mode, sec_attr, dispo, flags, template):
    hdl = FHANDLE(
        _CreateFile(
            file_name,
            access,
            share_mode,
            sec_attr,
            dispo,
            flags,
            template
            )
        )
    hdl.raise_on_invalid()
    return hdl

def create_file(
        file_name,
        access=GENERIC_READ | GENERIC_WRITE,
        share_mode=FILE_SHARE_READ | FILE_SHARE_WRITE,
        flags=FILE_ATTRIBUTE_NORMAL,
        dispo=OPEN_EXISTING,
        sec_attr=None,
        template=None
        ):
    return CreateFile(
        file_name,
        access,
        share_mode,
        sec_attr,
        dispo,
        flags,
        template
        )

################################################################################

_ReadFile = fun_fact(
    _k32.ReadFile, (BOOL, HANDLE, PVOID, DWORD, PDWORD, PVOID)
    )

def ReadFile(hdl, size_or_buf):
    buf_created = isinstance(size_or_buf, int)
    if buf_created:
        buf = byte_buffer(size_or_buf)
        size = size_or_buf
    else:
        buf = size_or_buf
        size = ctypes.sizeof(buf)

    num_read = DWORD()
    raise_on_zero(_ReadFile(hdl, buf, size, ref(num_read), None))
    if buf_created:
        return buf.raw[:num_read.value], num_read.value
    else:
        return buf, num_read.value

################################################################################

def read_file_text(hdl, size):
    buf, size = ReadFile(hdl, string_buffer(size))
    return buf[: size // WCHAR_SIZE]

################################################################################

_WriteFile = fun_fact(
    _k32.WriteFile,
    (BOOL, HANDLE, PVOID, DWORD, PDWORD, PVOID)
    )

def WriteFile(hdl, data):
    try:
        ldata = ctypes.sizeof(data)
        rdata = ref(data)
    except TypeError:
        rdata = byte_buffer(data)
        ldata = len(data)

    written = DWORD()
    raise_on_zero(_WriteFile(hdl, rdata, ldata, ref(written), None))
    return written.value

################################################################################

def write_file_text(hdl, txt):
    return WriteFile(hdl, string_buffer(txt, len(txt)))

################################################################################

_FlushFileBuffers = fun_fact(_k32.FlushFileBuffers, (BOOL, HANDLE))

def FlushFileBuffers(hdl):
    raise_on_zero(_FlushFileBuffers(hdl))

################################################################################

class _DUMMY_OVRLPD_STRUCT(ctypes.Structure):
    _fields_ = (
        ("Offset", DWORD),
        ("OffsetHigh", DWORD),
        )

class _DUMMY_OVRLPD_UNION(ctypes.Union):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("anon", _DUMMY_OVRLPD_STRUCT),
        ("Pointer", PVOID),
        )

class OVERLAPPED(ctypes.Structure):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("Internal", ULONG_PTR),
        ("InternalHigh", ULONG_PTR),
        ("anon", _DUMMY_OVRLPD_UNION),
        ("hEvent", HANDLE)
        )

POVERLAPPED = POINTER(OVERLAPPED)
PPOVERLAPPED = POINTER(POVERLAPPED)

################################################################################

_DeviceIoControl = fun_fact(
    _k32.DeviceIoControl, (
        BOOL,
        HANDLE,
        DWORD,
        PVOID,
        DWORD,
        PVOID,
        DWORD,
        PDWORD,
        POVERLAPPED
        )
    )

def DeviceIoControl(hdl, ioctl, in_ctobj, out_len):
    bytes_returned = DWORD(0)

    if in_ctobj is None or in_ctobj == 0:
        iptr, ilen = None, 0
    else:
        iptr, ilen = ref(in_ctobj), ctypes.sizeof(in_ctobj)

    if out_len is None or out_len == 0:
        out, optr, olen = None, None, 0
    else:
        out = byte_buffer(out_len)
        optr, olen = ref(out), out_len

    raise_on_zero(
        _DeviceIoControl(
            hdl,
            ioctl,
            iptr,
            ilen,
            optr,
            olen,
            ref(bytes_returned),
            None
            )
        )
    if num_ret := bytes_returned.value:
        return out if num_ret == olen else (CHAR * num_ret).from_buffer(out)
    return None

################################################################################

_CreateIoCompletionPort = fun_fact(
    _k32.CreateIoCompletionPort,
    (KHANDLE, HANDLE, HANDLE, ULONG_PTR, DWORD)
    )

def CreateIoCompletionPort(file, existing, key, num_threads):
    ioport = _CreateIoCompletionPort(file, existing, key, num_threads)
    ioport.raise_on_invalid()
    return ioport

def create_io_completion_port(file, key, num_threads=0, existing=None):
    return CreateIoCompletionPort(file, existing, key, num_threads)

################################################################################

_GetQueuedCompletionStatus = fun_fact(
    _k32.GetQueuedCompletionStatus,
    (BOOL, HANDLE, PDWORD, PULONG_PTR, PPOVERLAPPED, DWORD)
    )

def GetQueuedCompletionStatus(port, timeout):
    num_bytes = DWORD()
    key = ULONG_PTR()
    ovrl = POVERLAPPED()
    raise_on_zero(
        _GetQueuedCompletionStatus(
            port,
            ref(num_bytes),
            ref(key),
            ref(ovrl),
            timeout)
            )
    return num_bytes.value, key.value, ovrl

################################################################################

_CreateNamedPipe = fun_fact(
    _k32.CreateNamedPipeW, (
        HANDLE,
        PWSTR,
        DWORD,
        DWORD,
        DWORD,
        DWORD,
        DWORD,
        DWORD,
        PSECURITY_ATTRIBUTES,
        )
    )

def CreateNamedPipe(
        name,
        open_mode,
        pipe_mode,
        num_inst,
        out_buf_size,
        in_buf_bsize,
        default_timeout,
        p_sec_attr
        ):
    hdl = FHANDLE(
        _CreateNamedPipe(
            name,
            open_mode,
            pipe_mode,
            num_inst,
            out_buf_size,
            in_buf_bsize,
            default_timeout,
            p_sec_attr
            )
        )
    hdl.raise_on_invalid()
    return hdl

DEFAULT_PIPE_MODE = (
    PIPE_TYPE_MESSAGE |
    PIPE_READMODE_MESSAGE |
    PIPE_REJECT_REMOTE_CLIENTS
    )

def create_named_pipe(
        name,
        open_mode=PIPE_ACCESS_DUPLEX,
        pipe_mode=DEFAULT_PIPE_MODE,
        num_inst=PIPE_UNLIMITED_INSTANCES,
        out_buf_size=0x400,
        in_buf_bsize=0x400,
        default_timeout=0,
        p_sec_attr=None
        ):
    return CreateNamedPipe(
        name,
        open_mode,
        pipe_mode,
        num_inst,
        out_buf_size,
        in_buf_bsize,
        default_timeout,
        p_sec_attr
        )

################################################################################

_ConnectNamedPipe = fun_fact(_k32.ConnectNamedPipe, (BOOL, HANDLE, PVOID))

def ConnectNamedPipe(hdl):
    if not _ConnectNamedPipe(hdl, None):
        if (err := GetLastError()) == ERROR_PIPE_CONNECTED:
            return
        raise WinError(err)

################################################################################

_DisconnectNamedPipe = fun_fact(_k32.DisconnectNamedPipe, (BOOL, HANDLE))

def DisconnectNamedPipe(hdl):
    raise_on_zero(_DisconnectNamedPipe(hdl))

################################################################################

GetCurrentProcess = fun_fact(_k32.GetCurrentProcess, (HANDLE,))

################################################################################

GetCurrentProcessId = fun_fact(_k32.GetCurrentProcessId, (DWORD,))

################################################################################

GetCurrentThread = fun_fact(_k32.GetCurrentThread, (HANDLE,))

################################################################################

_ProcessIdToSessionId = fun_fact(
    _k32.ProcessIdToSessionId, (BOOL, DWORD, PDWORD)
    )

def ProcessIdToSessionId(pid):
    session = DWORD()
    raise_on_zero(_ProcessIdToSessionId(pid, ref(session)))
    return session.value

################################################################################

_GetModuleHandle = fun_fact(_k32.GetModuleHandleW, (HANDLE, PWSTR))

def GetModuleHandle(mod_name):
    res = _GetModuleHandle(mod_name)
    raise_on_zero(res)
    return res

################################################################################

_GetModuleFileName = fun_fact(
    _k32.GetModuleFileNameW, (DWORD, HANDLE, PWSTR, DWORD)
    )

def GetModuleFileName(hmod):
    size = 128
    res = size
    while res >= size:
        size *= 2
        buf = string_buffer(size)
        res = _GetModuleFileName(hmod, buf, size)
    raise_on_zero(res)
    return buf.value

################################################################################

_WaitForSingleObject = fun_fact(
    _k32.WaitForSingleObject, (DWORD, HANDLE, DWORD)
    )

def WaitForSingleObject(handle, timeout):
    res = _WaitForSingleObject(handle, timeout)
    raise_if(res == WAIT_FAILED)
    return res

################################################################################

_OpenProcess = fun_fact(
    _k32.OpenProcess, (HANDLE, DWORD, BOOL, DWORD)
    )

def OpenProcess(desired_acc, inherit, pid):
    res = KHANDLE(_OpenProcess(desired_acc, inherit, pid))
    res.raise_on_invalid()
    return res

################################################################################

_TerminateProcess = fun_fact(
    _k32.TerminateProcess, (BOOL, HANDLE, UINT)
    )

def TerminateProcess(handle, exit_code):
    raise_on_zero(_TerminateProcess(handle, exit_code))

################################################################################

_GetExitCodeProcess = fun_fact(
    _k32.GetExitCodeProcess, (BOOL, HANDLE, PDWORD)
    )

def GetExitCodeProcess(handle):
    exit_code = DWORD()
    raise_on_zero(_GetExitCodeProcess(handle, ref(exit_code)))
    return exit_code.value

################################################################################

_GetCommandLine = fun_fact(_k32.GetCommandLineW, (PWSTR,))
def GetCommandLine():
    return ctypes.wstring_at(_GetCommandLine())

################################################################################

_QueryDosDevice = fun_fact(
    _k32.QueryDosDeviceW, (DWORD, PWSTR, PWSTR, DWORD)
    )

def QueryDosDevice(device_name):
    size = 512
    buf = string_buffer(size)
    while True:
        if res := _QueryDosDevice(device_name, buf, size):
            if device_name is None:
                return multi_str_from_ubuf(buf, res)
            return buf.value[:res]
        raise_if(GetLastError() != ERROR_INSUFFICIENT_BUFFER)
        size *= 2
        buf = string_buffer(size)

################################################################################

def GetSystemTime():
    st = SYSTEMTIME()
    _k32.GetSystemTime(ref(st))
    return st

################################################################################

def GetSystemTimeAsFileTime():
    ft = FILETIME()
    _k32.GetSystemTimeAsFileTime(ref(ft))
    return ft

################################################################################

def SetSystemTime(st):
    raise_on_zero(_k32.SetSystemTime(ref(st)))

################################################################################

def GetLocalTime():
    st = SYSTEMTIME()
    _k32.GetLocalTime(ref(st))
    return st

################################################################################

def SetLocalTime(st):
    raise_on_zero(_k32.SetLocalTime(ref(st)))

################################################################################

def FileTimeToSystemTime(ft):
    st = SYSTEMTIME()
    raise_on_zero(_k32.FileTimeToSystemTime(ref(ft), ref(st)))
    return st

################################################################################

def SystemTimeToFileTime(st):
    ft = FILETIME()
    raise_on_zero(_k32.SystemTimeToFileTime(ref(st), ref(ft)))
    return ft

################################################################################

def FileTimeToLocalFileTime(ft):
    lft = FILETIME()
    raise_on_zero(_k32.FileTimeToLocalFileTime(ref(ft), ref(lft)))
    return lft

################################################################################

def FileTimeToLocalSystemTime(ft):
    st = FileTimeToSystemTime(ft)
    raise_on_zero(_k32.SystemTimeToTzSpecificLocalTime(0, ref(st), ref(st)))
    return st

################################################################################

def AdjustTime(seconds_to_adjust):
    ft = GetSystemTimeAsFileTime()
    ft += int(seconds_to_adjust * 1e7)
    st = FileTimeToSystemTime(ft)
    raise_on_zero(_k32.SetSystemTime(ref(st)))

################################################################################

def get_local_tzinfo():
    utc = GetSystemTimeAsFileTime()
    local = FileTimeToLocalFileTime(utc)
    return _dt.timezone(_dt.timedelta(0, int(local - utc) // 10_000_000))

################################################################################

def GetCurrentThreadId():
    return _k32.GetCurrentThreadId()

################################################################################

def GetFileAttributes(fname):
    res = _k32.GetFileAttributesW(fname)
    raise_if(res == INVALID_FILE_ATTRIBUTES)
    return res

################################################################################

_SetFileAttributes = fun_fact(
    _k32.SetFileAttributesW, (BOOL, PWSTR, DWORD)
    )

################################################################################

def SetFileAttributes(fname, attribs):
    raise_on_zero(_SetFileAttributes(fname, attribs))

################################################################################

_GetACP = fun_fact(_k32.GetACP, (DWORD,))

def GetACP():
    return _GetACP()

def get_ansi_encoding():
    return f"cp{GetACP()}"

################################################################################

_OutputDebugStringW = fun_fact(_k32.OutputDebugStringW, (None, PWSTR))

def OutputDebugString(dstr):
    _OutputDebugStringW(dstr)

################################################################################

def dbg_print(*args, end="\n"):
    _OutputDebugStringW(f"{' '.join(map(str, args))}{end}")

################################################################################

_SetThreadExecutionState = fun_fact(
    _k32.SetThreadExecutionState, (DWORD, DWORD)
    )

def SetThreadExecutionState(es_flags):
    return _SetThreadExecutionState(es_flags)

################################################################################

_GetPrivateProfileSectionNames = fun_fact(
    _k32.GetPrivateProfileSectionNamesW,
    (DWORD, PWSTR, DWORD, PWSTR)
    )

def GetPrivateProfileSectionNames(filename):
    size = 512
    buf = string_buffer(size)
    res = _GetPrivateProfileSectionNames(buf, size, filename)
    while res == size - 2:
        size *= 2
        buf = string_buffer(size)
        res = _GetPrivateProfileSectionNames(buf, size, filename)
    return multi_str_from_ubuf(buf, res)

################################################################################

_GetPrivateProfileSection = fun_fact(
    _k32.GetPrivateProfileSectionW,
    (DWORD, PWSTR, PWSTR, DWORD, PWSTR)
    )

def GetPrivateProfileSection(secname, filename):
    size = 512
    buf = string_buffer(size)
    res = _GetPrivateProfileSection(secname, buf, size, filename)
    while res == size - 2:
        size *= 2
        buf = string_buffer(size)
        res = _GetPrivateProfileSection(secname, buf, size, filename)
    entries = multi_str_from_ubuf(buf, res)
    d = _collections.OrderedDict()
    for e in entries:
        k, v = e.split("=", 1)
        d[k] = v
    return d

################################################################################

_WritePrivateProfileSection = fun_fact(
    _k32.WritePrivateProfileSectionW,
    (DWORD, PWSTR, PWSTR, PWSTR)
    )

def WritePrivateProfileSection(secname, secdata, filename):
    if not isinstance(secdata, str):
        lines = []
        for k, v in secdata.items():
            lines.append(f"{k}={v}")
        if lines:
            lines.append("\0")
            secdata = "\0".join(lines)
        else:
            secdata = "\0\0"
    raise_on_zero(_WritePrivateProfileSection(secname, secdata, filename))

################################################################################

_GetEnvironmentVariable = fun_fact(
    _k32.GetEnvironmentVariableW,
    (DWORD, PWSTR, PWSTR, DWORD)
    )

def GetEnvironmentVariable(name):
    size = 512
    while True:
        var = string_buffer(size)
        req = _GetEnvironmentVariable(name, var, size)
        raise_on_zero(req)
        if req <= size:
            break
        size = req
    return var.value

################################################################################

_SetEnvironmentVariable = fun_fact(
    _k32.SetEnvironmentVariableW,
    (BOOL, PWSTR, PWSTR)
    )

def SetEnvironmentVariable(name, value):
    raise_on_zero(_SetEnvironmentVariable(name, value))

################################################################################

# using void pointers instead of PWSTR so we can do pointer arithmatic.

_FreeEnvironmentStrings = fun_fact(
    _k32.FreeEnvironmentStringsW,
    (BOOL, PVOID)
    )

_GetEnvironmentStrings = fun_fact(_k32.GetEnvironmentStringsW, (PVOID,))

def GetEnvironmentStrings():
    ptr = _GetEnvironmentStrings()
    raise_on_zero(ptr)
    try:
        return multi_str_from_addr(ptr)
    finally:
        raise_on_zero(_FreeEnvironmentStrings(ptr))

def env_str_to_dict(estr):
    return dict(s.split("=", 1) for s in estr if s[0] != "=")

def get_env_as_dict():
    return env_str_to_dict(GetEnvironmentStrings())

################################################################################

_SetEnvironmentStrings = fun_fact(
    _k32.SetEnvironmentStringsW,
    (BOOL, PWSTR)
    )

def SetEnvironmentStrings(strings):
    raise_on_zero(_SetEnvironmentStrings(strings))

################################################################################

_ExpandEnvironmentStrings = fun_fact(
    _k32.ExpandEnvironmentStringsW,
    (DWORD, PWSTR, PWSTR, DWORD)
    )

def ExpandEnvironmentStrings(template):
    size = len(template)
    while True:
        var = string_buffer(size)
        req = _ExpandEnvironmentStrings(template, var, size)
        raise_on_zero(req)
        if req <= size:
            break
        size = req
    return var.value

################################################################################

class PROCESS_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("hProcess", KHANDLE),
        ("hThread", KHANDLE),
        ("dwProcessId", DWORD),
        ("dwThreadId", DWORD),
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.hThread.close()
        self.hProcess.close()

PPROCESS_INFORMATION = POINTER(PROCESS_INFORMATION)

################################################################################

class STARTUPINFO(ctypes.Structure):
    _fields_ = (
        ("cb", DWORD),
        ("lpReserved", PWSTR),
        ("lpDesktop", PWSTR),
        ("lpTitle", PWSTR),
        ("dwX", DWORD),
        ("dwY", DWORD),
        ("dwXSize", DWORD),
        ("dwYSize", DWORD),
        ("dwXCountChars", DWORD),
        ("dwYCountChars", DWORD),
        ("dwFillAttribute", DWORD),
        ("dwFlags", DWORD),
        ("wShowWindow", WORD),
        ("cbReserved2", WORD),
        ("lpReserved2", PBYTE),
        ("hStdInput", HANDLE),
        ("hStdOutput", HANDLE),
        ("hStdError", HANDLE),
        )

    def __init__(self):
        self.cb = ctypes.sizeof(STARTUPINFO)

PSTARTUPINFO = POINTER(STARTUPINFO)

class STARTUPINFOEX(ctypes.Structure):
    _fields_ = (
        ("StartupInfo", STARTUPINFO),
        ("lpAttributeList", PVOID),
        )

    def __init__(self, attr_lst=None):
        self.StartupInfo.cb = ctypes.sizeof(STARTUPINFOEX)
        self.lpAttributeList = attr_lst

################################################################################

_InitializeProcThreadAttributeList = fun_fact(
    _k32.InitializeProcThreadAttributeList,
    (BOOL, PVOID, DWORD, DWORD, PSIZE_T)
    )

def InitializeProcThreadAttributeList(alst, acnt, flags, size=0):
    size = SIZE_T(size)
    ok = _InitializeProcThreadAttributeList(alst, acnt, flags, ref(size))
    raise_if(not ok and alst)
    return size.value

################################################################################

_UpdateProcThreadAttribute = fun_fact(
    _k32.UpdateProcThreadAttribute,
    (BOOL, PVOID, DWORD, UINT_PTR, PVOID, SIZE_T, PVOID, PSIZE_T)
    )

def UpdateProcThreadAttribute(alst, flags, id, attr):
    raise_on_zero(
        _UpdateProcThreadAttribute(
            alst,
            flags,
            id,
            ref(attr),
            SIZE_T(ctypes.sizeof(attr)),
            None,
            None
            )
        )

################################################################################

_DeleteProcThreadAttributeList = fun_fact(
    _k32.DeleteProcThreadAttributeList, (None, PVOID)
    )

def DeleteProcThreadAttributeList(alst):
    _DeleteProcThreadAttributeList(alst)

################################################################################

class ProcThreadAttributeList:
    def __init__(self, attr_pairs):
        self.buf = None
        size = InitializeProcThreadAttributeList(None, len(attr_pairs), 0)
        buf = byte_buffer(size)
        InitializeProcThreadAttributeList(buf, 1, 0, size)
        try:
            for id, value in attr_pairs:
                UpdateProcThreadAttribute(buf, 0, id, value)
        except OSError:
            DeleteProcThreadAttributeList(buf)
            raise
        self.buf = buf

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.buf:
            DeleteProcThreadAttributeList(self.buf)
            self.buf = None

    def address(self):
        return ctypes.addressof(self.buf) if self.buf else None

################################################################################

_CreateProcess = fun_fact(
    _k32.CreateProcessW, (
        BOOL,
        PWSTR,
        PWSTR,
        PSECURITY_ATTRIBUTES,
        PSECURITY_ATTRIBUTES,
        BOOL,
        DWORD,
        PVOID,
        PWSTR,
        PSTARTUPINFO,
        PPROCESS_INFORMATION
        )
    )

def CreateProcess(
        app_name,
        cmd_line,
        proc_attr,
        thread_attr,
        inherit,
        cflags,
        env,
        curdir,
        startup_info,
        ):
    proc_info = PROCESS_INFORMATION()
    if isinstance(startup_info, STARTUPINFOEX):
        psi = ref(startup_info.StartupInfo)
    else:
        psi = ref(startup_info)
    raise_on_zero(
        _CreateProcess(
            app_name,
            cmd_line,
            ref(proc_attr) if proc_attr is not None else None,
            ref(thread_attr) if thread_attr is not None else None,
            inherit,
            cflags,
            env,
            curdir,
            psi,
            ref(proc_info)
            )
        )
    return proc_info

################################################################################

def create_process(
        arglist,
        cflags=0,
        startup_info=None,
        inherit=False,
        env=None,
        curdir=None,
        proc_attr=None,
        thread_attr=None,
        ):
    if startup_info is None:
        startup_info = STARTUPINFO()
    return CreateProcess(
        None,
        cmdline_from_args(arglist),
        proc_attr,
        thread_attr,
        inherit,
        cflags,
        env,
        curdir,
        startup_info
        )

################################################################################

_CreateJobObject = fun_fact(
    _k32.CreateJobObjectW,
    (KHANDLE, PSECURITY_ATTRIBUTES, PWSTR)
    )

def CreateJobObject(attrib=None, name=None):
    attrib = None if attrib is None else ref(attrib)
    job = _CreateJobObject(attrib, name)
    job.raise_on_invalid()
    return job

################################################################################

_AssignProcessToJobObject = fun_fact(
    _k32.AssignProcessToJobObject,
    (BOOL, HANDLE, HANDLE)
    )

def AssignProcessToJobObject(job, proc):
    raise_on_zero(_AssignProcessToJobObject(job, proc))

################################################################################

class JOBOBJECT_ASSOCIATE_COMPLETION_PORT(ctypes.Structure):
    _fields_ = (
        ("CompletionKey", PVOID),
        ("CompletionPort", HANDLE),
        )

class JOBOBJECT_BASIC_ACCOUNTING_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("TotalUserTime", LARGE_INTEGER),
        ("TotalKernelTime", LARGE_INTEGER),
        ("ThisPeriodTotalUserTime", LARGE_INTEGER),
        ("ThisPeriodTotalKernelTime", LARGE_INTEGER),
        ("TotalPageFaultCount", DWORD),
        ("TotalProcesses", DWORD),
        ("ActiveProcesses", DWORD),
        ("TotalTerminatedProcesses", DWORD),
    )

################################################################################

_SetInformationJobObject = fun_fact(
    _k32.SetInformationJobObject,
    (BOOL, HANDLE, INT, PVOID, DWORD)
    )

def SetInformationJobObject(job, cls, ct_info):
    size = ctypes.sizeof(ct_info)
    raise_on_zero(_SetInformationJobObject(job, cls, ref(ct_info), size))

################################################################################

_QueryInformationJobObject = fun_fact(
    _k32.QueryInformationJobObject,
    (BOOL, HANDLE, INT, PVOID, DWORD, PDWORD)
    )

def QueryInformationJobObject(job, cls, res_obj):
    raise_on_zero(
        _QueryInformationJobObject(
            job,
            cls,
            ref(res_obj),
            ctypes.sizeof(res_obj),
            None
            )
        )
    return res_obj

################################################################################

_ResumeThread = fun_fact(_k32.ResumeThread, (DWORD, HANDLE))

def ResumeThread(thdl):
    scnt = _ResumeThread(thdl)
    raise_if(scnt == 0xffffffff)
    return scnt

################################################################################

_GetPriorityClass = fun_fact(_k32.GetPriorityClass, (DWORD, HANDLE))

def GetPriorityClass(phdl):
    raise_on_zero(result := _GetPriorityClass(phdl))
    return result

################################################################################

_SetPriorityClass = fun_fact(_k32.SetPriorityClass, (BOOL, HANDLE, DWORD))

def SetPriorityClass(phdl, prio):
    raise_on_zero(_SetPriorityClass(phdl, prio))

################################################################################

_GetThreadPriority = fun_fact(_k32.GetThreadPriority, (INT, HANDLE))

def GetThreadPriority(thdl):
    result = _GetThreadPriority(thdl)
    raise_if(result == THREAD_PRIORITY_ERROR_RETURN)
    return result

################################################################################

_SetThreadPriority = fun_fact(_k32.SetThreadPriority, (BOOL, HANDLE, INT))

def SetThreadPriority(thdl, prio):
    raise_on_zero(_SetThreadPriority(thdl, prio))

################################################################################

def _get_dir(func, order):
    buf_size = 256
    while True:
        buf = string_buffer(buf_size)
        req_size = func(*((buf, buf_size) if order else (buf_size, buf)))
        if req_size <= buf_size:
            return buf.value
        buf_size = req_size

################################################################################

_GetSystemDirectory = fun_fact(_k32.GetSystemDirectoryW, (UINT, PWSTR, UINT))

def GetSystemDirectory():
    return _get_dir(_GetSystemDirectory, 1)

################################################################################

_GetWindowsDirectory = fun_fact(_k32.GetWindowsDirectoryW, (UINT, PWSTR, UINT))

def GetWindowsDirectory():
    return _get_dir(_GetWindowsDirectory, 1)


################################################################################

_GetCurrentDirectory = fun_fact(
    _k32.GetCurrentDirectoryW,
    (DWORD, DWORD, PWSTR)
    )

def GetCurrentDirectory():
    return _get_dir(_GetCurrentDirectory, 0)

################################################################################

_GetSystemWow64Directory = fun_fact(
    _k32.GetSystemWow64DirectoryW,
    (UINT, PWSTR, UINT)
    )

def GetSystemWow64Directory():
    return _get_dir(_GetSystemWow64Directory, 1)

################################################################################

_SetCurrentDirectory = fun_fact(_k32.SetCurrentDirectoryW, (BOOL, PWSTR))

def SetCurrentDirectory(path):
    raise_on_zero(_SetCurrentDirectory(path))

################################################################################

class ACTCTX(ctypes.Structure):
    _fields_ = (
        ("cbSize", ULONG),
        ("dwFlags", DWORD),
        ("lpSource", PWSTR),
        ("wProcessorArchitecture", USHORT),
        ("wLangId", USHORT),
        ("lpAssemblyDirectory", PWSTR),
        ("lpResourceName", PWSTR),
        ("lpApplicationName", PWSTR),
        ("hModule", HANDLE),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PACTCTX = POINTER(ACTCTX)

################################################################################

_CreateActCtx = fun_fact(_k32.CreateActCtxW, (HANDLE, PACTCTX))

def CreateActCtx(actctx):
    res = _CreateActCtx(ref(actctx))
    raise_if(res == INVALID_HANDLE_VALUE)
    return res

################################################################################

_ActivateActCtx = fun_fact(_k32.ActivateActCtx, (BOOL, HANDLE, PULONG_PTR))

def ActivateActCtx(ctx):
    cookie = ULONG_PTR()
    raise_on_zero(_ActivateActCtx(ctx, ref(cookie)))
    return cookie.value

################################################################################

_DeactivateActCtx = fun_fact(_k32.DeactivateActCtx, (BOOL, DWORD, ULONG_PTR))

def DeactivateActCtx(flags, cookie):
    raise_on_zero(_DeactivateActCtx(flags, cookie))

################################################################################

ReleaseActCtx = fun_fact(_k32.ReleaseActCtx, (None, HANDLE))

################################################################################

_GlobalAddAtom = fun_fact(_k32.GlobalAddAtomW, (WORD, PWSTR))

def GlobalAddAtom(name):
    atom = _GlobalAddAtom(name)
    raise_on_zero(atom)
    return atom

################################################################################

def global_add_atom(name):
    return ctypes.cast(GlobalAddAtom(name), PWSTR)

################################################################################

GlobalDeleteAtom = fun_fact(_k32.GlobalDeleteAtom, (None, WORD))

################################################################################

_GlobalGetAtomName = fun_fact(_k32.GlobalGetAtomNameW, (UINT, WORD, PWSTR, INT))

def GlobalGetAtomName(atom):
    size = 512
    while True:
        var = string_buffer(size)
        req = _GlobalGetAtomName(atom, var, size)
        raise_on_zero(req)
        if req <= size:
            break
        size = req
    return var.value

################################################################################

_FreeLibrary = fun_fact(_k32.FreeLibrary, (BOOL, HANDLE))

def FreeLibrary(hmod):
    raise_on_zero(_FreeLibrary(hmod))

################################################################################

class HMODULE(ScdToBeClosed, HANDLE, close_func=FreeLibrary, invalid=0):
    pass

################################################################################

_LoadLibraryEx = fun_fact(_k32.LoadLibraryExW, (HANDLE, PWSTR, HANDLE, DWORD))

def LoadLibraryEx(filename, flags=0):
    hmod = HMODULE(_LoadLibraryEx(filename, None, flags))
    hmod.raise_on_invalid()
    return hmod

def LoadLibrary(filename):
    return LoadLibraryEx(filename, 0)

################################################################################

class terminate_on_exception:

    def __enter__(self):
        return self

    def __exit__(self, typ, val, tb):
        if typ is None:
            return

        # An exception has occurred, which our caller would like to be taken
        # as a reason to terminate this process. Most likely our caller wants
        # this, because it is executing a callback from C code into python code
        # (through ctypes). In such a situation there is no possibility to
        # propagate this exception to the python interpreter and therefore the
        # process has to be terminated.
        # Before we do that, we try to inform the user.

        try:
            from ctwin32 import user  # noqa: PLC0415
            info = "".join(traceback.format_exception(typ, val, tb))
            try:
                interactive = user.is_interactive_process()
            except OSError:
                interactive = False

            if interactive:
                if sys.stderr is None or not hasattr(sys.stderr, "mode"):
                    user.txt_to_clip(info)
                    info += "\nThe above text has been copied to the clipboard."
                    user.MessageBox(
                        None,
                        info,
                        "Terminating program",
                        MB_OK | MB_ICONERROR
                        )
                else:
                    sys.stderr.write(info)
            else:
                dbg_print(info)

        finally:
            # Calling sys.exit() here won't help, since it depends on exception
            # propagation. We could hope that this thread is pumping messages
            # while watching for WM_QUIT messages and post such a message.
            # Since this possibility seems too vague, we play it safe
            # and call:
            ExitProcess(1)

################################################################################

_EnumResNameCallback = ctypes.WINFUNCTYPE(
    BOOL,
    HANDLE,
    PVOID,  # must not use PWSTR as parameter type since ctypes conversion
    PVOID,  # code will get confused by 16-bit IDs that are NOT valid pointers
    CallbackContextPtr
    )

@_EnumResNameCallback
def _EnumResNameCb(hmod, typ, name, ctxt):
    # cannot propagate exceptions from callback
    with terminate_on_exception():
        typ = typ if not (typ >> 16) else ctypes.wstring_at(typ)
        name = name if not (name >> 16) else ctypes.wstring_at(name)
        cbc = ctxt.contents
        res = cbc.callback(hmod, typ, name, cbc.context)
        # keep on enumerating if the callback fails to return a value
        return res if res is not None else True

_EnumResourceNames = fun_fact(
    _k32.EnumResourceNamesW,
    (BOOL, HANDLE, PWSTR, _EnumResNameCallback, CallbackContextPtr)
    )

def EnumResourceNames(hmod, typ, callback, context):
    cbc = CallbackContext(callback, context)
    if not _EnumResourceNames(hmod, typ, _EnumResNameCb, ref(cbc)):
        if (err := GetLastError()) != ERROR_RESOURCE_ENUM_USER_STOP:
            raise WinError(err)

################################################################################

def get_resource_names(hmod, typ):
    names = []

    @_EnumResNameCallback
    def collect(not_used1, not_used2, name, not_used3):
        # cannot propagate exceptions from callback
        with terminate_on_exception():
            if name >= 0x10000:
                name = PWSTR(name).value
            names.append(name)
            return True

    raise_on_zero(_EnumResourceNames(hmod, typ, collect, None))
    return names

################################################################################

_FindResource = fun_fact(
    _k32.FindResourceW, (HANDLE, HANDLE, PWSTR, PWSTR)
    )

def FindResource(hmod, name, typ):
    name = name if isinstance(name, PWSTR) else PWSTR(name)
    typ = typ if isinstance(typ, PWSTR) else PWSTR(typ)
    res = _FindResource(hmod, name, typ)
    raise_on_zero(res)
    return res

################################################################################

_SizeofResource = fun_fact(_k32.SizeofResource, (DWORD, HANDLE, HANDLE))

def SizeofResource(hmod, hrsc):
    res = _SizeofResource(hmod, hrsc)
    raise_on_zero(res)
    return res

################################################################################

_LoadResource = fun_fact(_k32.LoadResource, (PVOID, HANDLE, HANDLE))

def LoadResource(hmod, hrsc):
    res = _LoadResource(hmod, hrsc)
    raise_on_zero(res)
    return res

################################################################################

def get_resource_info(hmod, name, typ):
    hrsc = FindResource(hmod, name, typ)
    size = SizeofResource(hmod, hrsc)
    return LoadResource(hmod, hrsc), size

################################################################################

MESSAGE_RESOURCE_ANSI = 0
MESSAGE_RESOURCE_UNICODE = 1
MESSAGE_RESOURCE_UTF8 = 2

class MESSAGE_RESOURCE_ENTRY(ctypes.Structure):
    _fields_ = (
        ("Length", WORD),
        ("Flags", WORD),
        ("Text", BYTE),  # an array of (Length - offsetof(Text)) bytes
    )

class MESSAGE_RESOURCE_BLOCK(ctypes.Structure):
    _fields_ = (
        ("LowId", DWORD),
        ("HighId", DWORD),
        ("OffsetToEntries", DWORD),
    )

class MESSAGE_RESOURCE_DATA(ctypes.Structure):
    _fields_ = (
        ("NumberOfBlocks", DWORD),
        ("Blocks", MESSAGE_RESOURCE_BLOCK),  # an array of NumberOfBlocks
    )

def load_message_string(hmod, msg_id):
    msg_id = DWORD(msg_id).value
    for name in get_resource_names(hmod, RT_MESSAGETABLE):
        addr, _ = get_resource_info(hmod, name, RT_MESSAGETABLE)
        nblocks = MESSAGE_RESOURCE_DATA.from_address(addr).NumberOfBlocks
        bsize = ctypes.sizeof(MESSAGE_RESOURCE_BLOCK)
        bbase = addr + MESSAGE_RESOURCE_DATA.Blocks.offset
        for b in range(nblocks):
            block = MESSAGE_RESOURCE_BLOCK.from_address(bbase + b * bsize)
            if block.LowId <= msg_id <= block.HighId:
                eaddr = addr + block.OffsetToEntries
                for cur_id in range(block.LowId, block.HighId + 1):
                    entry = MESSAGE_RESOURCE_ENTRY.from_address(eaddr)
                    if cur_id != msg_id:
                        eaddr += entry.Length
                        continue
                    toff = MESSAGE_RESOURCE_ENTRY.Text.offset
                    taddr = eaddr + toff
                    slen = entry.Length - toff
                    if entry.Flags == MESSAGE_RESOURCE_UNICODE:
                        msg = ctypes.wstring_at(taddr, slen // 2)
                    else:
                        msg = ctypes.string_at(taddr, slen).decode(
                            "utf-8" if entry.Flags == MESSAGE_RESOURCE_UTF8
                            else get_ansi_encoding()
                            )
                    return msg.strip("\0")

    raise WinError(ERROR_RESOURCE_NAME_NOT_FOUND)

################################################################################

class SYSTEM_INFO(ctypes.Structure):
    _fields_ = (
        ("wProcessorArchitecture", WORD),
        ("wReserved", WORD),
        ("dwPageSize", DWORD),
        ("lpMinimumApplicationAddress", PVOID),
        ("lpMaximumApplicationAddress", PVOID),
        ("dwActiveProcessorMask", ULONG_PTR),
        ("dwNumberOfProcessors", DWORD),
        ("dwProcessorType", DWORD),
        ("dwAllocationGranularity", DWORD),
        ("wProcessorLevel", WORD),
        ("wProcessorRevision", WORD),
        )
PSYSTEM_INFO = POINTER(SYSTEM_INFO)

_GetSystemInfo = fun_fact(_k32.GetSystemInfo, (None, PSYSTEM_INFO))

def GetSystemInfo():
    si = SYSTEM_INFO()
    _GetSystemInfo(ref(si))
    return ns_from_struct(si)

################################################################################

_GetNativeSystemInfo = fun_fact(_k32.GetNativeSystemInfo, (None, PSYSTEM_INFO))

def GetNativeSystemInfo():
    si = SYSTEM_INFO()
    _GetNativeSystemInfo(ref(si))
    return ns_from_struct(si)

################################################################################

_IsWow64Process = fun_fact(_k32.IsWow64Process, (BOOL, HANDLE, PBOOL))

def IsWow64Process(hprocess):
    res = BOOL()
    raise_on_zero(_IsWow64Process(hprocess, ref(res)))
    return res != 0

################################################################################

def get_wow64_info(hprocess):
    mach = USHORT(IMAGE_FILE_MACHINE_UNKNOWN)
    proc = USHORT(IMAGE_FILE_MACHINE_UNKNOWN)
    try:
        _IsWow64Process2 = fun_fact(
            _k32.IsWow64Process2,
            (BOOL, HANDLE, PUSHORT, PUSHORT)
            )
    except AttributeError:
        # IsWow64Process2 not available
        if IsWow64Process(hprocess):
            mach = USHORT(IMAGE_FILE_MACHINE_AMD64)
        else:
            mach = USHORT(IMAGE_FILE_MACHINE_I386)
    else:
        raise_on_zero(_IsWow64Process2(hprocess, ref(proc), ref(mach)))
    return mach.value, proc.value

################################################################################

_GetStdHandle = fun_fact(_k32.GetStdHandle, (HANDLE, DWORD))

def GetStdHandle(nhdl):
    res = _GetStdHandle(nhdl)
    raise_if(res == INVALID_HANDLE_VALUE)
    return res

################################################################################

_GetFileType = fun_fact(_k32.GetFileType, (DWORD, HANDLE))

def GetFileType(hdl):
    res = _GetFileType(hdl)
    if res == FILE_TYPE_UNKNOWN:
        if (err := GetLastError()) != ERROR_SUCCESS:
            raise WinError(err)
    return res

################################################################################

_SetConsoleTextAttribute = fun_fact(
    _k32.SetConsoleTextAttribute,
    (BOOL, HANDLE, WORD)
    )

def SetConsoleTextAttribute(hcon, attr):
    raise_on_zero(_SetConsoleTextAttribute(hcon, attr))

################################################################################

class COORD(ctypes.Structure):
    _fields_ = (
        ("X", SHORT),
        ("Y", SHORT),
        )

class SMALL_RECT(ctypes.Structure):
    _fields_ = (
        ("Left", SHORT),
        ("Top", SHORT),
        ("Right", SHORT),
        ("Bottom", SHORT),
        )

class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
    _fields_ = (
        ("dwSize", COORD),
        ("dwCursorPosition", COORD),
        ("wAttributes", WORD),
        ("srWindow", SMALL_RECT),
        ("dwMaximumWindowSize", COORD),
        )
PCONSOLE_SCREEN_BUFFER_INFO = POINTER(CONSOLE_SCREEN_BUFFER_INFO)

################################################################################

_GetConsoleScreenBufferInfo = fun_fact(
    _k32.GetConsoleScreenBufferInfo,
    (BOOL, HANDLE, PCONSOLE_SCREEN_BUFFER_INFO)
    )

def GetConsoleScreenBufferInfo(hcon):
    info = CONSOLE_SCREEN_BUFFER_INFO()
    raise_on_zero(_GetConsoleScreenBufferInfo(hcon, ref(info)))
    return info

################################################################################

_FillConsoleOutputCharacter = fun_fact(
    _k32.FillConsoleOutputCharacterW,
    (BOOL, HANDLE, WCHAR, DWORD, COORD, PDWORD)
    )

def FillConsoleOutputCharacter(hdl, char, length, coord):
    num_written = DWORD()
    raise_on_zero(
        _FillConsoleOutputCharacter(hdl, char, length, coord, ref(num_written))
        )
    return num_written

################################################################################

_FillConsoleOutputAttribute = fun_fact(
    _k32.FillConsoleOutputCharacterW,
    (BOOL, HANDLE, WORD, DWORD, COORD, PDWORD)
    )

def FillConsoleOutputAttribute(hdl, attr, length, coord):
    num_written = DWORD()
    raise_on_zero(
        _FillConsoleOutputCharacter(hdl, attr, length, coord, ref(num_written))
        )
    return num_written

################################################################################

_SetConsoleCursorPosition = fun_fact(
    _k32.SetConsoleCursorPosition, (BOOL, HANDLE, COORD)
    )

def SetConsoleCursorPosition(hdl, coord):
    raise_on_zero(_SetConsoleCursorPosition(hdl, coord))

################################################################################

def clear_screen(hdl):
    info = GetConsoleScreenBufferInfo(hdl)
    size = info.dwMaximumWindowSize
    length = size.X * size.Y
    attr = info.wAttributes
    coord = COORD(0, 0)
    FillConsoleOutputAttribute(hdl, attr, length, coord)
    FillConsoleOutputCharacter(hdl, ord(" "), length, coord)
    SetConsoleCursorPosition(hdl, coord)

################################################################################

def cls(hdl=None):
    if hdl is None:
        hdl = GetStdHandle(STD_OUTPUT_HANDLE)
        if GetFileType(hdl) != FILE_TYPE_CHAR:
            return
    clear_screen(hdl)

################################################################################

_GetConsoleMode = fun_fact(
    _k32.GetConsoleMode, (BOOL, HANDLE, PDWORD)
    )

def GetConsoleMode(hdl):
    mode = DWORD()
    raise_on_zero(_GetConsoleMode(hdl, ref(mode)))
    return mode.value

################################################################################

_SetConsoleMode = fun_fact(
    _k32.SetConsoleMode, (BOOL, HANDLE, DWORD)
    )

def SetConsoleMode(hdl, mode):
    raise_on_zero(_SetConsoleMode(hdl, mode))

################################################################################

def enable_virt_term(hdl=None):
    if hdl is None:
        hdl = GetStdHandle(STD_OUTPUT_HANDLE)
        if GetFileType(hdl) != FILE_TYPE_CHAR:
            return
    mode = GetConsoleMode(hdl) | ENABLE_VIRTUAL_TERMINAL_PROCESSING
    SetConsoleMode(hdl, mode)

################################################################################

_GetNumberOfConsoleInputEvents  = fun_fact(
    _k32.GetNumberOfConsoleInputEvents, (BOOL, HANDLE, PDWORD)
    )

def GetNumberOfConsoleInputEvents(hdl):
    num = DWORD()
    raise_on_zero(_GetNumberOfConsoleInputEvents(hdl, ref(num)))
    return num.value

################################################################################

class FOCUS_EVENT_RECORD(ctypes.Structure):
    _fields_ = (
        ("bSetFocus", BOOL),
        )

class KEY_EVENT_RECORD(ctypes.Structure):
    _fields_ = (
        ("bKeyDown", BOOL),
        ("wRepeatCount", WORD),
        ("wVirtualKeyCode", WORD),
        ("wVirtualScanCode", WORD),
        ("UnicodeChar", WCHAR),
        ("dwControlKeyState", DWORD),
        )

class MENU_EVENT_RECORD(ctypes.Structure):
    _fields_ = (
        ("dwCommandId", UINT),
        )

class MOUSE_EVENT_RECORD(ctypes.Structure):
    _fields_ = (
        ("dwMousePosition", COORD),
        ("dwButtonState", DWORD),
        ("dwControlKeyState", DWORD),
        ("dwEventFlags", DWORD),
        )

class WINDOW_BUFFER_SIZE_RECORD(ctypes.Structure):
    _fields_ = (
        ("dwSize", COORD),
        )

class INPUT_RECORD_EVENT_UNION(ctypes.Union):
    _fields_ = (
        ("KeyEvent", KEY_EVENT_RECORD),
        ("MouseEvent", MOUSE_EVENT_RECORD),
        ("WindowBufferSizeEvent", WINDOW_BUFFER_SIZE_RECORD),
        ("MenuEvent", MENU_EVENT_RECORD),
        ("FocusEvent", FOCUS_EVENT_RECORD),
        )

class INPUT_RECORD(ctypes.Structure):
    _fields_ = (
        ("EventType", WORD),
        ("Event", INPUT_RECORD_EVENT_UNION),
        )

PINPUT_RECORD = POINTER(INPUT_RECORD)

################################################################################

_ReadConsoleInput  = fun_fact(
    _k32.ReadConsoleInputW , (BOOL, HANDLE, PINPUT_RECORD, DWORD, PDWORD)
    )

def ReadConsoleInput(hdl, num_records=1):
    num_records = max(num_records, 1)
    ir = (INPUT_RECORD * num_records)()
    num_read = DWORD()
    raise_on_zero(_ReadConsoleInput(hdl, ir, num_records, ref(num_read)))
    return ir[:num_read.value]

################################################################################

SetErrorMode = fun_fact(_k32.SetErrorMode, (UINT, UINT))

################################################################################

_SetThreadErrorMode = fun_fact(_k32.SetThreadErrorMode, (BOOL, DWORD, PDWORD))

def SetThreadErrorMode(mode):
    old = DWORD()
    raise_on_zero(_SetThreadErrorMode(mode, ref(old)))
    return old.value

################################################################################

_FindClose = fun_fact(_k32.FindClose, (BOOL, HANDLE))

def FindClose(hdl):
    raise_on_zero(_FindClose(hdl))

################################################################################

class FFHANDLE(
        ScdToBeClosed,
        HANDLE,
        close_func=FindClose,
        invalid=INVALID_HANDLE_VALUE
        ):
    pass

################################################################################

_FindFirstFile = fun_fact(
    _k32.FindFirstFileW, (HANDLE, PWSTR, PWIN32_FIND_DATA)
    )

def FindFirstFile(name, ignore_not_found=False):
    find_data = WIN32_FIND_DATA()
    hdl = FFHANDLE(_FindFirstFile(name, ref(find_data)))
    if not hdl.is_valid():
        err = GetLastError()
        if ignore_not_found and err == ERROR_FILE_NOT_FOUND:
            return None, None
        raise WinError(err)
    return hdl, find_data

################################################################################

_FindNextFile = fun_fact(
    _k32.FindNextFileW, (BOOL, HANDLE, PWIN32_FIND_DATA)
    )

def FindNextFile(hdl):
    find_data = WIN32_FIND_DATA()
    res = _FindNextFile(hdl, ref(find_data))
    if not res:
        err = GetLastError()
        if err == ERROR_NO_MORE_FILES:
            return None
        raise WinError(err)
    return find_data

################################################################################

def iter_dir(directory, ignore_access_denied=True):
    pattern = str(directory) + "\\*"
    try:
        hdl, info = FindFirstFile(pattern, True)
    except OSError as e:
        if ignore_access_denied and e.winerror == ERROR_ACCESS_DENIED:
            return
        raise
    if hdl is None:
        return
    with hdl:
        while True:
            yield directory, info
            is_sub = (
                bool(info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) and
                info.cFileName not in (".", "..")
                )
            if is_sub:
                sub = rf"{directory}\{info.cFileName}"
                yield from iter_dir(sub, ignore_access_denied)

            info = FindNextFile(hdl)
            if info is None:
                break

################################################################################

def find_file(name):
    hdl, info = FindFirstFile(name)
    hdl.close()
    return info

################################################################################

_FindFirstFileName = fun_fact(
    _k32.FindFirstFileNameW, (HANDLE, PWSTR, DWORD, PDWORD, PWSTR)
    )

def FindFirstFileName(name):
    size = DWORD(256)
    while True:
        size.value *= 2
        buf = string_buffer(size.value)
        hdl = FFHANDLE(_FindFirstFileName(name, 0, ref(size), buf))
        if not hdl.is_valid():
            err = GetLastError()
            if err == ERROR_MORE_DATA:
                continue
            if err == ERROR_FILE_NOT_FOUND:
                return None, None
            raise WinError(err)
        return hdl, buf.value

################################################################################

_FindNextFileName = fun_fact(
    _k32.FindNextFileNameW, (BOOL, HANDLE, PDWORD, PWSTR)
    )

def FindNextFileName(hdl):
    size = DWORD(256)
    while True:
        size.value *= 2
        buf = string_buffer(size.value)
        if not _FindNextFileName(hdl, ref(size), buf):
            err = GetLastError()
            if err == ERROR_MORE_DATA:
                continue
            if err == ERROR_HANDLE_EOF:
                return None
            raise WinError(err)
        return buf.value

################################################################################

def find_all_filenames(name):
    hdl, name = FindFirstFileName(name)
    if hdl is not None:
        with hdl:
            while True:
                yield name
                name = FindNextFileName(hdl)
                if name is None:
                    break

################################################################################

class PowerRequest(_int_enum):
    DisplayRequired = 0
    SystemRequired = 1
    AwayModeRequired = 2
    ExecutionRequired = 3
    Inactive = -1

    @classmethod
    def from_param(cls, obj):
        return int(cls(obj))

# values for REASON_CONTEXT.Version and REASON_CONTEXT.Flags
POWER_REQUEST_CONTEXT_VERSION = 0
POWER_REQUEST_CONTEXT_SIMPLE_STRING = 1
POWER_REQUEST_CONTEXT_DETAILED_STRING = 2

class REASON_CONTEXT_DETAILED(ctypes.Structure):
    _fields_ = (
        ("LocalizedReasonModule", HMODULE),
        ("LocalizedReasonId", ULONG),
        ("ReasonStringCount", ULONG),
        ("ReasonStrings", PPWSTR),
        )

class REASON_CONTEXT_UNION(ctypes.Union):
    _fields_ = (
        ("Detailed", REASON_CONTEXT_DETAILED),
        ("SimpleReasonString", PWSTR),
        )

class REASON_CONTEXT(ctypes.Structure):
    _fields_ = (
        ("Version", ULONG),
        ("Flags", DWORD),
        ("Reason", REASON_CONTEXT_UNION)
        )

    def __init__(self, reason_str=""):
        self.Version = POWER_REQUEST_CONTEXT_VERSION
        self.Flags = POWER_REQUEST_CONTEXT_SIMPLE_STRING
        self.Reason = REASON_CONTEXT_UNION(SimpleReasonString=reason_str)

PREASON_CONTEXT = POINTER(REASON_CONTEXT)

################################################################################

_PowerCreateRequest = fun_fact(
    _k32.PowerCreateRequest, (HANDLE, PREASON_CONTEXT)
    )

def PowerCreateRequest(reason):
    hdl = FHANDLE(_PowerCreateRequest(ref(reason)))
    hdl.raise_on_invalid()
    return hdl

################################################################################

_PowerSetRequest = fun_fact(_k32.PowerSetRequest, (BOOL, HANDLE, PowerRequest))

def PowerSetRequest(hdl, pwr_request):
    raise_on_zero(_PowerSetRequest(hdl, pwr_request))

################################################################################

_PowerClearRequest = fun_fact(
    _k32.PowerClearRequest, (BOOL, HANDLE, PowerRequest)
    )

def PowerClearRequest(hdl, pwr_request):
    raise_on_zero(_PowerClearRequest(hdl, pwr_request))

################################################################################

def create_power_request(reason_str, pwr_request=PowerRequest.Inactive):
    hdl = PowerCreateRequest(REASON_CONTEXT(reason_str))
    if pwr_request != PowerRequest.Inactive:
        PowerSetRequest(hdl, pwr_request)
    return hdl

################################################################################

GetDriveType = fun_fact(_k32.GetDriveTypeW, (UINT, PWSTR))

################################################################################

_GetLogicalDriveStrings = fun_fact(
    _k32.GetLogicalDriveStringsW, (DWORD, DWORD, PWSTR)
    )

def GetLogicalDriveStrings():
    size = 256
    buf = string_buffer(size)
    raise_on_zero(res := _GetLogicalDriveStrings(size, buf))
    return multi_str_from_ubuf(buf, res)

################################################################################

_FindFirstVolume = fun_fact(_k32.FindFirstVolumeW, (HANDLE, PWSTR, DWORD))

def FindFirstVolume():
    size = 256
    buf = string_buffer(size)
    hdl = _FindFirstVolume(buf, size)
    raise_if(hdl == INVALID_HANDLE_VALUE)
    return hdl, buf.value

################################################################################

_FindNextVolume = fun_fact(_k32.FindNextVolumeW, (BOOL, HANDLE, PWSTR, DWORD))

def FindNextVolume(hdl):
    size = 256
    buf = string_buffer(size)
    raise_on_zero(_FindNextVolume(hdl, buf, size))
    return buf.value

################################################################################

_FindVolumeClose = fun_fact(_k32.FindVolumeClose, (BOOL, HANDLE))

def FindVolumeClose(hdl):
    raise_on_zero(_FindVolumeClose(hdl))

################################################################################

def enum_volumes():
    hdl, vol = FindFirstVolume()
    try:
        while True:
            yield vol
            vol = FindNextVolume(hdl)
    except OSError as e:
        if e.winerror != ERROR_NO_MORE_FILES:
            raise
    finally:
        FindVolumeClose(hdl)

################################################################################

_GetVolumePathNamesForVolumeName = fun_fact(
    _k32.GetVolumePathNamesForVolumeNameW, (BOOL, PWSTR, PWSTR, DWORD, PDWORD)
    )

def GetVolumePathNamesForVolumeName(vol):
    size = DWORD(256)
    while True:
        buf = string_buffer(size.value)
        ok = _GetVolumePathNamesForVolumeName(vol, buf, size, ref(size))
        if ok:
            return multi_str_from_ubuf(buf, size.value)
        elif (err := GetLastError()) != ERROR_MORE_DATA:
            raise_on_err(err)

################################################################################

_ReadProcessMemory = fun_fact(
    _k32.ReadProcessMemory, (BOOL, HANDLE, PVOID, PVOID, SIZE_T, PSIZE_T)
    )

def ReadProcessMemory(hdl, addr, length):
    buf = byte_buffer(length)
    raise_on_zero(_ReadProcessMemory(hdl, addr, buf, length, None))
    return buf

################################################################################

_WriteProcessMemory = fun_fact(
    _k32.WriteProcessMemory, (BOOL, HANDLE, PVOID, PVOID, SIZE_T, PSIZE_T)
    )

def WriteProcessMemory(hdl, addr, data):
    raise_on_zero(_WriteProcessMemory(hdl, addr, data, len(data), None))

################################################################################

def get_proc_env_as_dict(hdl):
    t32 = (    DWORD, 16,  72,  656)    # noqa: E201, RUF100
    t64 = (ULONGLONG, 32, 128, 1008)
    peb = ntdll.get_wow64_proc_env_blk(hdl)
    if peb != 0:
        # type, offs process params, offs env ptr, offs env len
        typ, opp, oep, oel = t32
    else:
        typ, opp, oep, oel = t64 if ctypes.sizeof(PVOID) == 8 else t32
        peb = ntdll.get_proc_env_blk(hdl)
    isize = ctypes.sizeof(typ)

    def read_proc_int(addr):
        return typ.from_buffer(ReadProcessMemory(hdl, addr, isize)).value

    process_params = read_proc_int(peb + opp)
    env_ptr = read_proc_int(process_params + oep)
    env_len = read_proc_int(process_params + oel)

    buf = ReadProcessMemory(hdl, env_ptr, env_len)
    str_len = env_len // ctypes.sizeof(WCHAR)
    str_type = WCHAR * str_len
    env = multi_str_from_ubuf(str_type.from_buffer(buf), str_len)
    return env_str_to_dict(env)

################################################################################

class BY_HANDLE_FILE_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("dwFileAttributes", DWORD),
        ("ftCreationTime", FILETIME),
        ("ftLastAccessTime", FILETIME),
        ("ftLastWriteTime", FILETIME),
        ("dwVolumeSerialNumber", DWORD),
        ("nFileSizeHigh", DWORD),
        ("nFileSizeLow", DWORD),
        ("nNumberOfLinks", DWORD),
        ("nFileIndexHigh", DWORD),
        ("nFileIndexLow", DWORD),
        )
PBY_HANDLE_FILE_INFORMATION = POINTER(BY_HANDLE_FILE_INFORMATION)

_GetFileInformationByHandle = fun_fact(
    _k32.GetFileInformationByHandle, (BOOL, HANDLE, PBY_HANDLE_FILE_INFORMATION)
    )

def GetFileInformationByHandle(hdl):
    info = BY_HANDLE_FILE_INFORMATION()
    raise_on_zero(_GetFileInformationByHandle(hdl, ref(info)))
    return ns_from_struct(info)

################################################################################

_CreateFileMapping = fun_fact(
    _k32.CreateFileMappingW, (
        HANDLE,
        HANDLE,
        PSECURITY_ATTRIBUTES,
        DWORD,
        DWORD,
        DWORD,
        PWSTR
        )
    )

def CreateFileMapping(fhdl, sec_attr, prot, maxsize, name=None):
    hdl = KHANDLE(
        _CreateFileMapping(
            fhdl,
            sec_attr,
            prot,
            maxsize >> 32,
            maxsize & 0xffffffff,
            name
            )
        )
    hdl.raise_on_invalid()
    return hdl

################################################################################

_MapViewOfFile = fun_fact(
    _k32.MapViewOfFile, (PVOID, HANDLE, DWORD, DWORD, DWORD, SIZE_T)
    )

def MapViewOfFile(mapping, acc, offset, size):
    addr = _MapViewOfFile(mapping, acc, offset >> 32, offset & 0xffffffff, size)
    raise_on_zero(addr)
    return addr

################################################################################

_UnmapViewOfFile = fun_fact(_k32.UnmapViewOfFile, (BOOL, PVOID))

def UnmapViewOfFile(addr):
    raise_on_zero(_UnmapViewOfFile(addr))

################################################################################

_GetFileSizeEx = fun_fact(_k32.GetFileSizeEx, (BOOL, HANDLE, PLARGE_INTEGER))

def GetFileSizeEx(hdl):
    size = LARGE_INTEGER()
    raise_on_zero(_GetFileSizeEx(hdl, ref(size)))
    return size.value

GetFileSize = GetFileSizeEx

################################################################################

_AddDllDirectory = fun_fact(_k32.AddDllDirectory, (PVOID, PWSTR))

def AddDllDirectory(dir_name):
    cookie = _AddDllDirectory(dir_name)
    raise_if(cookie is None)
    return cookie

################################################################################

_RemoveDllDirectory = fun_fact(_k32.RemoveDllDirectory, (BOOL, PVOID))

def RemoveDllDirectory(cookie):
    raise_on_zero(_RemoveDllDirectory(cookie))

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# Note: This sample will only do something meaningful if there is a
# Basler-USB-camera (https://www.baslerweb.com/) connected and the related
# driver installed.

import sys
import time
import msvcrt
from ctwin32 import setupapi, shell, advapi

################################################################################

def clear_keys():
    # Remove any keys that are present in the input queue.
    while msvcrt.kbhit():
        msvcrt.getch()

################################################################################

def main():
    print("press any key to stop...\n")

    guid = setupapi.SetupDiClassGuidsFromNameEx("PylonUSB")[0]

    # In a loop keep disabling and enabling all devices that belong to the
    # above device class until a key is pressed.

    clear_keys()
    while not msvcrt.kbhit():
        setupapi.disable_devices(guid)
        print("disabled")
        time.sleep(0.3)
        setupapi.enable_devices(guid)
        print("enabled")
        # Enabling is faster. So we wait a bit longer to get a duty cycle of
        # approx. 50%.
        time.sleep(1)

    clear_keys()

################################################################################

if __name__ == "__main__":

    # Enabling and disabling devices requires administrative privileges.
    if not advapi.running_as_admin():
        shell.elevate(sys.executable, __file__)
    else:
        main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from types import SimpleNamespace as _namespace
import ctypes
from . import (
    ERROR_INVALID_PARAMETER,
    fun_fact,
    ns_from_struct,
    ref,
    suppress_winerr,
    )
from .wtypes import (
    CHAR,
    FILETIME,
    HANDLE,
    INT,
    LARGE_INTEGER,
    LONG,
    LUID,
    NTSTATUS,
    PHANDLE,
    PLONG,
    PLUID,
    POINTER,
    PPLUID,
    PPVOID,
    PSTR,
    PVOID,
    PULONG,
    SIZE_T,
    ULONG,
    UNICODE_STRING,
    WORD,
    )
from .ntdll import raise_failed_status
from .kernel import (
    FileTimeToLocalFileTime,
    FileTimeToSystemTime,
    KHANDLE,
    get_ansi_encoding,
    get_local_tzinfo,
    )
from .advapi import (
    AllocateLocallyUniqueId,
    GetLengthSid,
    ConvertSidToStringSid,
    )

_sec = ctypes.WinDLL("secur32.dll", use_last_error=True)

################################################################################

class LSA_LAST_INTER_LOGON_INFO(ctypes.Structure):
    _fields_ = (
        ("LastSuccessfulLogon", LARGE_INTEGER),
        ("LastFailedLogon", LARGE_INTEGER),
        ("FailedAttemptCountSinceLastSuccessfulLogon", ULONG),
        )

class SECURITY_LOGON_SESSION_DATA(ctypes.Structure):
    _fields_ = (
        ("Size", ULONG),
        ("LogonId", LUID),
        ("UserName", UNICODE_STRING),
        ("LogonDomain", UNICODE_STRING),
        ("AuthenticationPackage", UNICODE_STRING),
        ("LogonType", ULONG),
        ("Session", ULONG),
        ("Sid", PVOID),
        ("LogonTime", LARGE_INTEGER),
        ("LogonServer", UNICODE_STRING),
        ("DnsDomainName", UNICODE_STRING),
        ("Upn", UNICODE_STRING),
        ("UserFlags", ULONG),
        ("LastLogonInfo", LSA_LAST_INTER_LOGON_INFO),
        ("LogonScript", UNICODE_STRING),
        ("ProfilePath", UNICODE_STRING),
        ("HomeDirectory", UNICODE_STRING),
        ("HomeDirectoryDrive", UNICODE_STRING),
        ("LogoffTime", LARGE_INTEGER),
        ("KickOffTime", LARGE_INTEGER),
        ("PasswordLastSet", LARGE_INTEGER),
        ("PasswordCanChange", LARGE_INTEGER),
        ("PasswordMustChange", LARGE_INTEGER),
        )
PSECURITY_LOGON_SESSION_DATA = POINTER(SECURITY_LOGON_SESSION_DATA)
PPSECURITY_LOGON_SESSION_DATA = POINTER(PSECURITY_LOGON_SESSION_DATA)

################################################################################

_LsaFreeReturnBuffer = fun_fact(
    _sec.LsaFreeReturnBuffer, (NTSTATUS, PVOID)
    )

def LsaFreeReturnBuffer(ptr):
    raise_failed_status(_LsaFreeReturnBuffer(ptr))

################################################################################

_LsaGetLogonSessionData = fun_fact(
    _sec.LsaGetLogonSessionData,
    (NTSTATUS, PLUID, PPSECURITY_LOGON_SESSION_DATA)
    )

def LsaGetLogonSessionData(luid):
    ltz = get_local_tzinfo()
    ptr = PSECURITY_LOGON_SESSION_DATA()
    raise_failed_status(_LsaGetLogonSessionData(ref(luid), ref(ptr)))
    try:
        lsd = ptr.contents
        lli = lsd.LastLogonInfo

        def la2dt(la):
            st = FileTimeToSystemTime(FILETIME(0))
            with suppress_winerr(ERROR_INVALID_PARAMETER):
                #convert utc to local time
                st = FileTimeToSystemTime(FileTimeToLocalFileTime(FILETIME(la)))
            return st.to_datetime(tzinfo=ltz)

        return _namespace(
            LogonId=int(lsd.LogonId),
            UserName=str(lsd.UserName),
            LogonDomain=str(lsd.LogonDomain),
            AuthenticationPackage=str(lsd.AuthenticationPackage),
            LogonType=lsd.LogonType,
            Session=lsd.Session,
            Sid=ConvertSidToStringSid(
                ctypes.string_at(lsd.Sid, GetLengthSid(lsd.Sid))
                ) if lsd.Sid else None,
            LogonTime=la2dt(lsd.LogonTime),
            LogonServer=str(lsd.LogonServer),
            DnsDomainName=str(lsd.DnsDomainName),
            Upn=str(lsd.Upn),
            UserFlags=lsd.UserFlags,
            LastLogonInfo=_namespace(
                LastSuccessfulLogon=la2dt(lli.LastSuccessfulLogon),
                LastFailedLogon=la2dt(lli.LastFailedLogon),
                FailedAttemptCountSinceLastSuccessfulLogon=(
                    lli.FailedAttemptCountSinceLastSuccessfulLogon
                    )
                ),
            LogonScript=str(lsd.LogonScript),
            ProfilePath=str(lsd.ProfilePath),
            HomeDirectory=str(lsd.HomeDirectory),
            HomeDirectoryDrive=str(lsd.HomeDirectoryDrive),
            LogoffTime=la2dt(lsd.LogoffTime),
            KickOffTime=la2dt(lsd.KickOffTime),
            PasswordLastSet=la2dt(lsd.PasswordLastSet),
            PasswordCanChange=la2dt(lsd.PasswordCanChange),
            PasswordMustChange=la2dt(lsd.PasswordMustChange)
            )
    finally:
        LsaFreeReturnBuffer(ptr)

################################################################################

_LsaEnumerateLogonSessions = fun_fact(
    _sec.LsaEnumerateLogonSessions, (NTSTATUS, PULONG, PPLUID)
    )

def LsaEnumerateLogonSessions():
    count = ULONG()
    pluid = PLUID()
    raise_failed_status(_LsaEnumerateLogonSessions(ref(count), ref(pluid)))
    try:
        return [
            LsaGetLogonSessionData(pluid[idx])
            for idx in reversed(range(count.value))
            ]
    finally:
        LsaFreeReturnBuffer(pluid)

################################################################################

_LsaDeregisterLogonProcess = fun_fact(
    _sec.LsaDeregisterLogonProcess, (NTSTATUS, HANDLE)
    )

def LsaDeregisterLogonProcess(hdl):
    raise_failed_status(_LsaDeregisterLogonProcess(hdl))

################################################################################

_LsaConnectUntrusted = fun_fact(_sec.LsaConnectUntrusted, (NTSTATUS, PHANDLE))

def LsaConnectUntrusted():
    hdl = HANDLE()
    raise_failed_status(_LsaConnectUntrusted(ref(hdl)))
    return hdl

################################################################################

class LSA_STRING(ctypes.Structure):
    _fields_ = (
        ("Length", WORD),
        ("MaximumLength", WORD),
        ("Buffer", PSTR),
        )

    def __str__(self):
        return ctypes.string_at(
            self.Buffer,
            self.Length
            ).decode(get_ansi_encoding())

PLSA_STRING = POINTER(LSA_STRING)

def LsaStrFromStr(init):
    if isinstance(init, str):
        init = init.encode(get_ansi_encoding())

    class SELF_CONTAINED_LSAS(ctypes.Structure):
        _fields_ = (
            ("ls", LSA_STRING),
            ("buf", CHAR * (1 + len(init))),
            )

        def __init__(self, init):
            li = len(init)
            baddr = ctypes.addressof(self) + __class__.buf.offset
            super().__init__((li, 1 + li, baddr), init)

        @property
        def ptr(self):
            return PLSA_STRING(self.ls)

    return SELF_CONTAINED_LSAS(init)

################################################################################

_LsaLookupAuthenticationPackage = fun_fact(
    _sec.LsaLookupAuthenticationPackage, (NTSTATUS, HANDLE, PLSA_STRING, PULONG)
    )

def LsaLookupAuthenticationPackage(hlsa, name):
    name = LsaStrFromStr(name.encode(get_ansi_encoding()))
    auth_pkg = ULONG()
    raise_failed_status(
        _LsaLookupAuthenticationPackage(
            hlsa,
            name.ptr,
            ref(auth_pkg)
            )
        )
    return auth_pkg.value

################################################################################

TOKEN_SOURCE_LENGTH = 8

class TOKEN_SOURCE(ctypes.Structure):
    _fields_ = (
        ("SourceName", CHAR * TOKEN_SOURCE_LENGTH),
        ("SourceIdentifier", LUID),
        )

PTOKEN_SOURCE = POINTER(TOKEN_SOURCE)

################################################################################

class QUOTA_LIMITS(ctypes.Structure):
    _fields_ = (
        ("PagedPoolLimit", SIZE_T),
        ("NonPagedPoolLimit", SIZE_T),
        ("MinimumWorkingSetSize", SIZE_T),
        ("MaximumWorkingSetSize", SIZE_T),
        ("PagefileLimit", SIZE_T),
        ("TimeLimit", LARGE_INTEGER),
        )

PQUOTA_LIMITS = POINTER(QUOTA_LIMITS)

################################################################################

_LsaLogonUser = fun_fact(
    _sec.LsaLogonUser, (
        NTSTATUS,
        HANDLE,
        PLSA_STRING,
        INT,
        ULONG,
        PVOID,
        ULONG,
        PVOID,
        PTOKEN_SOURCE,
        PPVOID,
        PULONG,
        PLUID,
        PHANDLE,
        PQUOTA_LIMITS,
        PLONG
        )
    )

def LsaLogonUser(
        lsa_hdl,
        origin_name,
        logon_type,
        auth_pkg,
        auth_info,
        local_groups=None,
        src_name=None,
        ):

    origin_name = LsaStrFromStr(origin_name)
    pgroups = None
    if local_groups is not None:
        pgroups = ref(local_groups)
    if src_name is None:
        src_name = "ctwin32"
    token_src = TOKEN_SOURCE(
        src_name.encode(get_ansi_encoding()),
        LUID(AllocateLocallyUniqueId())
        )

    profile = PVOID()
    plen = ULONG()
    lid = LUID()
    tok = KHANDLE()
    qlim = QUOTA_LIMITS()
    subs = LONG(0)

    status = _LsaLogonUser(
        lsa_hdl,
        origin_name.ptr,
        logon_type,
        auth_pkg,
        ref(auth_info),
        ctypes.sizeof(auth_info),
        pgroups,
        ref(token_src),
        ref(profile),
        ref(plen),
        ref(lid),
        ref(tok),
        ref(qlim),
        ref(subs)
        )
    if status < 0:
        raise_failed_status(subs if subs.value < 0 else status)

    try:
        sundries = _namespace(
            ProfileBuffer=ctypes.string_at(profile.value, plen.value),
            LogonId=int(lid),
            Quotas=ns_from_struct(qlim)
            )
        return tok, sundries
    finally:
        LsaFreeReturnBuffer(profile)

################################################################################
